import json
import logging
import shutil
import tempfile
import time
from copy import deepcopy
from pathlib import Path
from typing import Callable, Optional

from atopile import layout
from atopile.cli.logging import LoggingStage
from atopile.config import config
from atopile.errors import UserException, UserPickError
from atopile.front_end import DeprecatedException
from faebryk.core.cpp import set_max_paths
from faebryk.core.module import Module
from faebryk.core.pathfinder import MAX_PATHS
from faebryk.core.solver.defaultsolver import DefaultSolver
from faebryk.core.solver.nullsolver import NullSolver
from faebryk.core.solver.solver import Solver
from faebryk.exporters.bom.jlcpcb import write_bom_jlcpcb
from faebryk.exporters.netlist.graph import (
    attach_net_names,
    attach_nets,
)
from faebryk.exporters.parameters.parameters_to_file import export_parameters_to_file
from faebryk.exporters.pcb.kicad.artifacts import (
    export_dxf,
    export_gerber,
    export_glb,
    export_pick_and_place,
    export_step,
    githash_layout,
)
from faebryk.exporters.pcb.kicad.pcb import LibNotInTable, get_footprint
from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
from faebryk.exporters.pcb.pick_and_place.jlcpcb import (
    convert_kicad_pick_and_place_to_jlcpcb,
)
from faebryk.exporters.pcb.testpoints.testpoints import export_testpoints
from faebryk.library import _F as F
from faebryk.libs.app.checks import (
    DrcException,
    RequiresExternalUsageNotFulfilled,
    run_post_build_checks,
    run_post_pcb_checks,
    run_pre_build_checks,
)
from faebryk.libs.app.designators import (
    attach_random_designators,
    load_designators,
)
from faebryk.libs.app.pcb import (
    apply_layouts,
    apply_routing,
    check_net_names,
    create_footprint_library,
    ensure_footprint_lib,
    load_net_names,
)
from faebryk.libs.app.picking import load_descriptive_properties, save_parameters
from faebryk.libs.exceptions import (
    UserResourceException,
    accumulate,
    downgrade,
    iter_leaf_exceptions,
    iter_through_errors,
)
from faebryk.libs.kicad.fileformats_latest import C_kicad_fp_lib_table_file
from faebryk.libs.kicad.fileformats_version import try_load_kicad_pcb_file
from faebryk.libs.picker.picker import PickError, pick_part_recursively
from faebryk.libs.util import ConfigFlag, KeyErrorAmbiguous

logger = logging.getLogger(__name__)

SKIP_SOLVING = ConfigFlag("SKIP_SOLVING", default=False)


def _get_solver() -> Solver:
    if SKIP_SOLVING:
        logger.warning("Assertion checking is disabled")
        return NullSolver()
    else:
        return DefaultSolver()


def build(app: Module) -> None:
    """Build the project."""

    def G():
        return app.get_graph()

    solver = _get_solver()

    # TODO remove hack
    # Disables children pathfinding
    # ```
    # power1.lv ~ power2.lv
    # power1.hv ~ power2.hv
    # -> power1 is not connected power2
    # ```
    set_max_paths(int(MAX_PATHS), 0, 0)

    with LoggingStage("prebuild", "Running pre-build checks"):
        if not SKIP_SOLVING:
            logger.info("Resolving bus parameters")
            try:
                F.is_bus_parameter.resolve_bus_parameters(G())
            # FIXME: this is a hack around a compiler bug
            except KeyErrorAmbiguous as ex:
                raise UserException(
                    "Unfortunately, there's a compiler bug at the moment that means "
                    "that this sometimes fails. Try again, and it'll probably work. "
                    "See https://github.com/atopile/atopile/issues/807"
                ) from ex
        else:
            logger.warning("Skipping bus parameter resolution")

        logger.info("Running checks")
        try:
            run_pre_build_checks(app, G())
        except RequiresExternalUsageNotFulfilled as ex:
            # TODO ato code
            raise UserException(str(ex)) from ex

        # Pre-pick project checks - things to look at before time is spend ---------
        # Make sure the footprint libraries we're looking for exist
        consolidate_footprints(app)

    with LoggingStage("load-pcb", "Loading PCB"):
        pcb = try_load_kicad_pcb_file(config.build.paths.layout)
        transformer = PCB_Transformer(pcb.kicad_pcb, G(), app)
        if config.build.keep_designators:
            load_designators(G(), attach=True)

    with LoggingStage("picker", "Picking components") as stage:
        if config.build.keep_picked_parts:
            load_descriptive_properties(G())
        try:
            pick_part_recursively(app, solver, progress=stage)
        except* PickError as ex:
            raise ExceptionGroup(
                "Failed to pick parts for some modules",
                [UserPickError(str(e)) for e in iter_leaf_exceptions(ex)],
            ) from ex
        save_parameters(G())

    with LoggingStage("footprints", "Handling footprints"):
        # Use standard footprints for known packages regardless of
        # what's otherwise been specified.
        # FIXME: this currently includes explicitly set footprints, but shouldn't
        F.has_package.standardize_footprints(app, solver)
        create_footprint_library(app)

    with LoggingStage("nets", "Preparing nets"):
        attach_random_designators(G())
        nets = attach_nets(G())
        # We have to re-attach the footprints, and subsequently nets, because the first
        # attachment is typically done before the footprints have been created
        # and therefore many nets won't be re-attached properly. Also, we just created
        # and attached them to the design above, so they weren't even there to attach
        transformer.attach()
        if config.build.keep_net_names:
            load_net_names(G())
        attach_net_names(nets)
        check_net_names(G())

    with LoggingStage("postbuild", "Running post-build checks"):
        logger.info("Running checks")
        run_post_build_checks(app, G())

    with LoggingStage("update-pcb", "Updating PCB"):
        original_pcb = deepcopy(pcb)
        transformer.apply_design(config.build.paths.fp_lib_table)
        transformer.check_unattached_fps()

        if transform_trait := app.try_get_trait(F.has_layout_transform):
            logger.info("Transforming PCB")
            transform_trait.transform(transformer)

        # set layout
        apply_layouts(app)
        transformer.move_footprints()
        apply_routing(app, transformer)

        if pcb == original_pcb:
            if config.build.frozen:
                logger.info("No changes to layout. Passed --frozen check.")
            else:
                logger.info(
                    f"No changes to layout. Not writing {config.build.paths.layout}"
                )
        elif config.build.frozen:
            original_path = config.build.paths.output_base.with_suffix(
                ".original.kicad_pcb"
            )
            updated_path = config.build.paths.output_base.with_suffix(
                ".updated.kicad_pcb"
            )
            original_pcb.dumps(original_path)
            pcb.dumps(updated_path)

            # TODO: make this a real util
            def _try_relative(path: Path) -> Path:
                try:
                    return path.relative_to(Path.cwd(), walk_up=True)
                except ValueError:
                    return path

            original_relative = _try_relative(original_path)
            updated_relative = _try_relative(updated_path)

            raise UserException(
                "Built as frozen, but layout changed. \n"
                f"Original layout: {original_relative}\n"
                f"Updated layout: {updated_relative}\n"
                "You can see the changes by running:\n"
                f'`diff --color "{original_relative}" "{updated_relative}"`',
                title="Frozen failed",
                # No markdown=False here because we have both a command and paths
            )
        else:
            backup_file = config.build.paths.output_base.with_suffix(
                f".{time.strftime('%Y%m%d-%H%M%S')}.kicad_pcb"
            )
            logger.info(f"Backing up layout to {backup_file}")
            with config.build.paths.layout.open("rb") as f:
                backup_file.write_bytes(f.read())

            logger.info(f"Updating layout {config.build.paths.layout}")
            pcb.dumps(config.build.paths.layout)

    # Figure out what targets to build
    if config.build.targets == ["__default__"]:
        targets = muster.do_by_default
    elif config.build.targets == ["*"] or config.build.targets == ["all"]:
        targets = list(muster.targets.keys())
    else:
        targets = config.build.targets

    # Remove targets we don't know about, or are excluded
    excluded_targets = set(config.build.exclude_targets)
    known_targets = set(muster.targets.keys())
    targets = list(set(targets) - excluded_targets & known_targets)

    if generate_manufacturing_data.__muster_name__ in targets:  # type: ignore
        with LoggingStage("pre-manufacturing", "Running pre-manufacturing checks"):
            try:
                run_post_pcb_checks(app, G(), config.build.paths.layout)
            except DrcException as ex:
                raise UserException(f"Detected DRC violations: \n{ex.pretty()}") from ex

    # Make the noise
    built_targets = []
    with accumulate() as accumulator:
        for target_name in targets:
            with LoggingStage(
                f"target-{target_name}", f"Building [green]'{target_name}'[/green]"
            ):
                with accumulator.collect():
                    muster.targets[target_name](app, solver)
                built_targets.append(target_name)


TargetType = Callable[[Module, Solver], None]


class Muster:
    """A class to register targets to."""

    def __init__(self, logger: Optional[logging.Logger] = None) -> None:
        self.targets: dict[str, TargetType] = {}
        self.do_by_default: list[str] = []
        self.log = logger or logging.getLogger(__name__)

    def add_target(
        self, func: TargetType, name: Optional[str] = None, default: bool = True
    ):
        """Register a function as a target."""
        name = name or func.__name__
        self.targets[name] = func
        if default:
            self.do_by_default.append(name)
        return func

    def register(self, name: Optional[str] = None, default: bool = True):
        """Register a target under a given name."""

        def decorator(func: TargetType):
            self.add_target(func, name, default)
            func.__muster_name__ = name or func.__name__  # type: ignore
            return func

        return decorator


muster = Muster()


@muster.register("bom")
def generate_bom(app: Module, solver: Solver) -> None:
    """Generate a BOM for the project."""
    write_bom_jlcpcb(
        app.get_children_modules(types=Module),
        config.build.paths.output_base.with_suffix(".bom.csv"),
    )


@muster.register("mfg-data", default=False)
def generate_manufacturing_data(app: Module, solver: Solver) -> None:
    """
    Generate manufacturing artifacts for the project.
    - STEP
    - GLB
    - DXF
    - Gerber zip
    - Pick and place (default and JLCPCB)
    - Testpoint-location
    """
    # Create temp copy of layout file with git hash substituted
    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_layout = githash_layout(
            config.build.paths.layout,
            Path(tmpdir) / config.build.paths.layout.name,
        )

        export_step(
            tmp_layout,
            step_file=config.build.paths.output_base.with_suffix(".pcba.step"),
        )
        export_glb(
            tmp_layout,
            glb_file=config.build.paths.output_base.with_suffix(".pcba.glb"),
        )
        export_dxf(
            tmp_layout,
            dxf_file=config.build.paths.output_base.with_suffix(".pcba.dxf"),
        )

        export_gerber(
            tmp_layout,
            gerber_zip_file=config.build.paths.output_base.with_suffix(".gerber.zip"),
        )

        pnp_file = config.build.paths.output_base.with_suffix(".pick_and_place.csv")
        export_pick_and_place(tmp_layout, pick_and_place_file=pnp_file)
        convert_kicad_pick_and_place_to_jlcpcb(
            pnp_file,
            config.build.paths.output_base.with_suffix(".jlcpcb_pick_and_place.csv"),
        )

        export_testpoints(
            app,
            testpoints_file=config.build.paths.output_base.with_suffix(
                ".testpoints.json"
            ),
        )


@muster.register("manifest")
def generate_manifest(app: Module, solver: Solver) -> None:
    """Generate a manifest for the project."""
    with accumulate() as accumulator:
        with accumulator.collect():
            manifest = {}
            manifest["version"] = "2.0"
            for build in config.builds:
                with build:
                    if config.build.paths.layout:
                        by_layout_manifest = manifest.setdefault(
                            "by-layout", {}
                        ).setdefault(str(config.build.paths.layout), {})
                        by_layout_manifest["layouts"] = str(
                            config.build.paths.output_base.with_suffix(".layouts.json")
                        )

            manifest_path = config.project.paths.manifest
            manifest_path.parent.mkdir(exist_ok=True, parents=True)
            with open(manifest_path, "w", encoding="utf-8") as f:
                json.dump(manifest, f, indent=4)


@muster.register("layout-module-map")
def generate_module_map(app: Module, solver: Solver) -> None:
    """Generate a designator map for the project."""
    layout.generate_module_map(app)


@muster.register("variable-report")
def generate_variable_report(app: Module, solver: Solver) -> None:
    """Generate a report of all the variable values in the design."""
    # TODO: support other file formats
    export_parameters_to_file(
        app, solver, config.build.paths.output_base.with_suffix(".variables.md")
    )


def consolidate_footprints(app: Module) -> None:
    """
    Consolidate all the project's footprints into a single directory.

    TODO: @v0.4 remove this, it's a fallback for v0.2 designs
    If there's an entry named "lib" pointing at "build/footprints/footprints.pretty"
    then copy all footprints we can find there
    """
    fp_ids_to_check = []
    for fp_t in app.get_children(False, types=(F.has_footprint)):
        fp = fp_t.get_footprint()
        if has_identifier_t := fp.try_get_trait(F.KicadFootprint.has_kicad_identifier):
            fp_ids_to_check.append(has_identifier_t.kicad_identifier)

    try:
        fptable = C_kicad_fp_lib_table_file.loads(config.build.paths.fp_lib_table)
    except FileNotFoundError:
        fptable = C_kicad_fp_lib_table_file.skeleton()

    # TODO: @windows might need to check backslashes
    lib_in_fptable = any(
        lib.name == "lib" and lib.uri.endswith("build/footprints/footprints.pretty")
        for lib in fptable.fp_lib_table.libs
    )
    lib_prefix_on_ids = any(fp_id.startswith("lib:") for fp_id in fp_ids_to_check)
    if not lib_in_fptable and not lib_prefix_on_ids:
        # no "lib" entry pointing to the footprints dir, this project isn't broken
        return
    elif lib_prefix_on_ids and not lib_in_fptable:
        # we need to add a lib entry pointing to the footprints dir
        ensure_footprint_lib(
            "lib",
            config.project.paths.build / "footprints" / "footprints.pretty",
            fptable,
        )
    elif lib_in_fptable and not lib_prefix_on_ids:
        # We could probably just remove the lib entry
        # but let's be conservative for now
        logger.info(
            "It seems like this project is using a legacy footprint consolidation "
            "unnecessarily. You can likely remove the 'lib' entry from the "
            "'fp-lib-table' file."
        )

    fp_target = config.project.paths.build / "footprints" / "footprints.pretty"
    fp_target_step = config.project.paths.build / "footprints" / "footprints.3dshapes"
    fp_target.mkdir(exist_ok=True, parents=True)

    if not config.project.paths.root:
        with downgrade(UserResourceException):
            raise UserResourceException(
                "No project root directory found. Cannot consolidate footprints."
            )
        return

    for fp in config.project.paths.root.glob("**/*.kicad_mod"):
        try:
            shutil.copy(fp, fp_target)
        except shutil.SameFileError:
            logger.debug("Footprint '%s' already exists in the target directory", fp)

    # Post-process all the footprints in the target directory
    for fp in fp_target.glob("**/*.kicad_mod"):
        with open(fp, "r+", encoding="utf-8") as file:
            content = file.read()
            content = content.replace("{build_dir}", str(fp_target_step))
            file.seek(0)
            file.write(content)
            file.truncate()

    # TODO: @v0.4 increase the level of this to WARNING
    # when there's an alternative
    with downgrade(DeprecatedException, to_level=logging.DEBUG):
        raise DeprecatedException(
            "This project uses a deprecated footprint consolidation mechanism."
        )

    # Finally, check that we have all the footprints we know we will need
    try:
        for err_collector, fp_id in iter_through_errors(fp_ids_to_check):
            with err_collector():
                get_footprint(fp_id, config.build.paths.fp_lib_table)
    except* (FileNotFoundError, LibNotInTable) as ex:

        def _make_user_resource_exception(e: Exception) -> UserResourceException:
            if isinstance(e, FileNotFoundError):
                return UserResourceException(
                    f"Footprint library {e.filename} doesn't exist"
                )
            elif isinstance(e, LibNotInTable):
                return UserResourceException(
                    f"Footprint library {e.lib_id} not found in {e.lib_table_path}"
                )
            assert False, "How'd we get here?"

        raise ex.derive(
            [_make_user_resource_exception(e) for e in iter_leaf_exceptions(ex)]
        ) from ex



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod

import faebryk.library._F as F


class has_equal_pins(F.Footprint.TraitT):
    @abstractmethod
    def get_pin_map(self) -> dict[F.Pad, str]: ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod

import faebryk.library._F as F


class can_attach_via_pinmap(F.Footprint.TraitT):
    @abstractmethod
    def attach(self, pinmap: dict[str, F.Electrical | None]): ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod

from faebryk.core.parameter import Parameter


class is_representable_by_single_value(Parameter.TraitT):
    @abstractmethod
    def get_single_representing_value(self): ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import Enum, auto

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module, ModuleException
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P
from faebryk.libs.util import assert_once

logger = logging.getLogger(__name__)


class CH342(Module):
    """
    Base class for CH342x USB to double UART converter
    """

    class IntegratedLDO(Module):
        power_in: F.ElectricPower
        power_out: F.ElectricPower

        def __preinit__(self):
            F.ElectricLogic.connect_all_module_references(self, gnd_only=True)

            self.power_out.voltage.constrain_superset(
                L.Range.from_center_rel(3.3 * P.V, 0.1)
            )
            self.power_in.voltage.constrain_subset(L.Range(4 * P.V, 5.5 * P.V))

        @L.rt_field
        def bridge(self):
            return F.can_bridge_defined(self.power_in, self.power_out)

    class IOPowerConfiguration(Enum):
        USB_5V = auto()
        """IO powered by USB 5V"""
        INTERNAL_3V3 = auto()
        """IO powered by the integrated 3.3V regulator"""
        EXTERNAL = auto()
        """IO powered by an external 1.8V-5V source"""

    class ChipPowerConfiguration(Enum):
        USB_5V = auto()
        """Chip powered by USB 5V"""
        EXTERNAL_5V = auto()
        """Chip powered by an external 5V source"""
        EXTERNAL_3V3 = auto()
        """Chip powered by an external 3.3V source"""

    @assert_once
    def set_power_configuration(
        self,
        chip_power_configuration: ChipPowerConfiguration = ChipPowerConfiguration.USB_5V,  # noqa: E501
        io_voltage_configuration: IOPowerConfiguration = IOPowerConfiguration.INTERNAL_3V3,  # noqa: E501
    ):
        """Configure how the chip is powered, and what the io voltage will be."""
        # how is the chip powered
        if chip_power_configuration == self.ChipPowerConfiguration.EXTERNAL_3V3:
            # short the integrated regulator power input to the output to disable
            self.integrated_regulator.power_in.connect(self.power_3v)
        elif chip_power_configuration == self.ChipPowerConfiguration.USB_5V:
            # use the USB power to power the chip
            self.usb.usb_if.buspower.connect(self.integrated_regulator.power_in)
        else:
            # use an external 5V power source for the chip
            ...

        # how is the IO powered
        if io_voltage_configuration == self.IOPowerConfiguration.INTERNAL_3V3:
            # check if the integrated regulator is not disabled
            if chip_power_configuration == self.ChipPowerConfiguration.EXTERNAL_3V3:
                raise ModuleException(
                    self,
                    "Cannot power IO from the integrated regulator when it is disabled, use 'EXTERNAL' power configuration instead",  # noqa: E501
                )
            # io is 3v3 and powered by the integrated regulator (10mA max)
            self.power_io.connect(self.power_3v)
            F.ElectricLogic.connect_all_module_references(
                self,
                exclude={
                    self.integrated_regulator.power_in,
                },
            )
        elif io_voltage_configuration == self.IOPowerConfiguration.USB_5V:
            # io is 5v and powered by USB
            self.power_io.connect(self.usb.usb_if.buspower)
            F.ElectricLogic.connect_all_module_references(
                self,
                exclude={
                    self.integrated_regulator.power_in,
                    self.integrated_regulator.power_out,
                    self.power_3v,
                },
            )
        else:
            # io is 1.8V-5V and powered by an external source
            ...

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    usb: F.USB2_0
    # uart_base = L.list_field(2, F.UART_Base)

    integrated_regulator: IntegratedLDO
    power_io: F.ElectricPower
    power_3v: F.ElectricPower

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://wch-ic.com/downloads/CH342DS1_PDF.html"
    )
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    def __preinit__(self):
        # ----------------------------------------
        #                aliasess
        # ----------------------------------------
        # ----------------------------------------
        #            parametrization
        # ----------------------------------------
        self.power_3v.voltage.constrain_subset(L.Range.from_center_rel(3.3 * P.V, 0.1))
        self.power_io.voltage.constrain_subset(L.Range(1.7 * P.V, 5.5 * P.V))

        # ----------------------------------------
        #              connections
        # ----------------------------------------
        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)

        # chip internal connection
        self.integrated_regulator.power_out.connect(self.power_3v)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class pf_74AHCT2G125(Module):
    """
    The 74AHC1G/AHCT1G125 is a high-speed Si-gate CMOS device.
    The 74AHC1G/AHCT1G125 provides one non-inverting buffer/line
    driver with 3-state output. The 3-state output is controlled
    by the output enable input (OE). A HIGH at OE causes the
    output to assume a high-impedance OFF-state.
    """

    # interfaces

    power: F.ElectricPower
    a: F.ElectricLogic  # IN
    y: F.ElectricLogic  # OUT
    oe: F.ElectricLogic  # enable, active low

    @L.rt_field
    def attach_to_footprint(self):
        x = self
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": x.oe.line,
                "2": x.a.line,
                "3": x.power.lv,
                "4": x.y.line,
                "5": x.power.hv,
            }
        )

    def __preinit__(self):
        self.power.voltage.constrain_subset(L.Range(4.5 * P.V, 5.5 * P.V))

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.a, self.y)

    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://datasheet.lcsc.com/lcsc/2304140030_Nexperia-74AHCT1G125GV-125_C12494.pdf"
    )

    @L.rt_field
    def can_be_decoupled(self):
        class _(F.can_be_decoupled.impl()):
            def decouple(self, owner: Module):
                obj = self.get_obj(pf_74AHCT2G125)
                obj.power.decoupled.decouple(owner=owner)

        return _()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import json
import logging
from pathlib import Path
from typing import Any, Callable, Iterable

from atopile.errors import UserBadParameterError
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.core.parameter import Expression, Is, Parameter, Predicate
from faebryk.core.solver.solver import Solver
from faebryk.libs.sets.sets import P_Set
from faebryk.libs.util import EquivalenceClasses, groupby, ind, typename

logger = logging.getLogger(__name__)


def parameter_alias_classes(G: Graph) -> list[set[Parameter]]:
    full_eq = EquivalenceClasses[Parameter](GraphFunctions(G).nodes_of_type(Parameter))

    is_exprs = [e for e in GraphFunctions(G).nodes_of_type(Is) if e.constrained]

    for is_expr in is_exprs:
        params_ops = [op for op in is_expr.operands if isinstance(op, Parameter)]
        full_eq.add_eq(*params_ops)

    return full_eq.get()


def get_params_for_expr(expr: Expression) -> set[Parameter]:
    param_ops = {op for op in expr.operatable_operands if isinstance(op, Parameter)}
    expr_ops = {op for op in expr.operatable_operands if isinstance(op, Expression)}

    return param_ops | {op for e in expr_ops for op in get_params_for_expr(e)}


def parameter_dependency_classes(G: Graph) -> list[set[Parameter]]:
    related = EquivalenceClasses[Parameter](GraphFunctions(G).nodes_of_type(Parameter))

    eq_exprs = [e for e in GraphFunctions(G).nodes_of_type(Predicate) if e.constrained]

    for eq_expr in eq_exprs:
        params = get_params_for_expr(eq_expr)
        related.add_eq(*params)

    return related.get()


def parameter_report(G: Graph, path: Path):
    params = GraphFunctions(G).nodes_of_type(Parameter)
    exprs = GraphFunctions(G).nodes_of_type(Expression)
    predicates = {e for e in exprs if isinstance(e, Predicate)}
    exprs.difference_update(predicates)
    alias_classes = parameter_alias_classes(G)
    eq_classes = parameter_dependency_classes(G)
    unused = [
        p
        for p in params
        if not any(isinstance(e.node, Expression) for e in p.operated_on.edges)
    ]

    def non_empty(classes: list[set[Parameter]]):
        return [c for c in classes if len(c) > 1]

    def bound(classes: list[set[Parameter]]):
        return sum(len(c) for c in non_empty(classes))

    infostr = (
        f"{len(params)} parameters"
        f"\n    {len(non_empty(alias_classes))}({bound(alias_classes)}) alias classes"
        f"\n    {len(non_empty(eq_classes))}({bound(eq_classes)}) equivalence classes"
        f"\n    {len(unused)} unused"
        "\n"
    )
    infostr += f"{len(exprs)} expressions, {len(predicates)} predicates"

    logger.info(f"Found {infostr}")

    out = ""
    out += infostr + "\n"

    def block(
        header: str,
        f: Callable[[], str] | None = None,
        lines: list[str] | list[list[str]] | None = None,
    ):
        nonlocal out
        out_str = ""
        if f:
            out_str += f()
        if lines:
            lines = [n for n in lines if isinstance(n, str)] + [
                n for nested in lines if isinstance(nested, list) for n in nested
            ]
            out_str += "\n".join(lines)

        out += f"{header}{'-' * 80}\n{ind(out_str)}\n"

    block(
        "Parameters",
        lines=sorted([p.get_full_name(types=True) for p in params]),
    )

    block(
        "Unused",
        lines=sorted([p.get_full_name(types=True) for p in unused]),
    )

    def Eq(classes: list[set[Parameter]]):
        stream = ""
        for eq_class in classes:
            if len(eq_class) <= 1:
                continue
            stream += "\n    ".join(
                sorted([p.get_full_name(types=True) for p in eq_class])
            )
            stream += "\n"
        return stream.removesuffix("\n")

    block(
        "Fully aliased",
        f=lambda: Eq(alias_classes),
    )

    block(
        "Equivalence classes",
        f=lambda: Eq(eq_classes),
    )

    def type_group(name: str, container: Iterable):
        type_grouped = sorted(
            groupby(container, lambda x: type(x)).items(), key=lambda x: typename(x[0])
        )
        block(
            name,
            lines=[
                f"{typename(type_)}: {len(list(group))}"
                for type_, group in type_grouped
            ],
        )

    type_group("Expressions", exprs)
    type_group("Predicates", predicates)

    path.write_text(out, encoding="utf-8")


def _generate_json_parameters(parameters: dict[str, dict[str, P_Set[Any]]]) -> str:
    json_parameters = {
        module_name: {
            param_name: str(param_value)
            for param_name, param_value in module_params.items()
        }
        for module_name, module_params in parameters.items()
        if module_params
    }

    return json.dumps(json_parameters, indent=2)


def _generate_md_parameters(parameters: dict[str, dict[str, P_Set[Any]]]) -> str:
    out = "# Module Parameters\n"
    out += "| Module | Parameter | Value |\n"
    out += "| --- | --- | --- |\n"
    for module_name, paras in sorted(parameters.items()):
        if paras:
            # Sort parameters for consistent output
            sorted_params = sorted(paras.items())
            # First parameter of the module shows the module name
            if sorted_params:
                par_name, par_value = sorted_params[0]
                out += (
                    f"| `{module_name.replace('|', '\\|')}` | "
                    f"`{par_name.replace('|', '\\|')}` | "
                    f"`{str(par_value).replace('|', '\\|')}` |\n"
                )
                # Subsequent parameters of the same module have empty module cell
                for par_name, par_value in sorted_params[1:]:
                    out += (
                        f"|  | "
                        f"`{par_name.replace('|', '\\|')}` | "
                        f"`{str(par_value).replace('|', '\\|')}` |\n"
                    )
    out += "\n"

    return out


def _generate_txt_parameters(parameters: dict[str, dict[str, P_Set[Any]]]) -> str:
    out = ""
    for module_name, paras in sorted(parameters.items()):
        if paras:
            out += f"{module_name}\n"
            out += "\n".join(
                [
                    f"    {par_name}: {par_value}\n"
                    for par_name, par_value in paras.items()
                ]
            )
            out += "\n"

    return out


def export_parameters_to_file(module: Module, solver: Solver, path: Path):
    """Write all parameters of the given module to a file."""
    # {module_name: [{param_name: param_value}, {param_name: param_value},...]}

    parameters = dict[str, dict[str, P_Set[Any]]]()

    for m in module.get_children_modules(types=Module):
        module_name = m.get_full_name(types=True)
        module_params = m.get_children(direct_only=True, types=Parameter)
        param_names = [param.get_full_name().split(".")[-1] for param in module_params]
        param_values = [
            solver.inspect_get_known_supersets(param) for param in module_params
        ]
        parameters[module_name] = {
            name: value for name, value in zip(param_names, param_values)
        }

    logger.info(f"Writing parameters to {path}")

    match path.suffix:
        case ".txt":
            out = _generate_txt_parameters(parameters)
        case ".md":
            out = _generate_md_parameters(parameters)
        case ".json":
            out = _generate_json_parameters(parameters)
        case _:
            raise UserBadParameterError(
                f"Export to file extension [{path.suffix}] not supported in {path}"
            )

    path.parent.mkdir(parents=True, exist_ok=True)
    temp_path = path.with_suffix(path.suffix + ".tmp")

    try:
        temp_path.write_text(out, encoding="utf-8")
        temp_path.replace(path)
    finally:
        if temp_path.exists():
            temp_path.unlink()



import logging
from typing import Annotated, Iterator

import typer
from git import Repo
from semver import Version

import faebryk.libs.backend.packages.api as packages_api
from atopile.config import config
from atopile.errors import UserBadParameterError, UserException
from faebryk.libs.backend.packages.api import PackagesAPIClient
from faebryk.libs.package.dist import Dist, DistValidationError

# Set up logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

package_app = typer.Typer(rich_markup_mode="rich")


FROM_GIT = "from-git"


def _yield_semver_tags() -> Iterator[Version]:
    repo = Repo(config.project.paths.root)
    for tag in repo.tags:
        if not tag.commit == repo.head.commit:
            continue

        try:
            yield Version.parse(tag.name.removeprefix("v"))
        except ValueError:
            continue


def _apply_version(specd_version: str) -> None:
    if specd_version == FROM_GIT:
        semver_tags = list(_yield_semver_tags())

        if len(semver_tags) == 0:
            raise UserBadParameterError("No semver tags found for the current commit")

        elif len(semver_tags) > 1:
            raise UserBadParameterError(
                "Multiple semver tags found for the current commit: %s."
                " No guessing which to use."
            )

        version = semver_tags[0]

    else:
        try:
            version = Version.parse(specd_version.removeprefix("v"))
        except ValueError as ex:
            raise UserBadParameterError(
                f"{specd_version} is not a valid semantic version"
            ) from ex

    if version.prerelease or version.build:
        raise UserBadParameterError(
            "Version must be a semantic version, without prerelease or build."
            f" Got {str(version)}"
        )

    assert config.project.package is not None
    config.project.package.version = str(version)


@package_app.command()
def publish(
    version: Annotated[
        str,
        typer.Option(
            "--version",
            "-v",
            envvar="ATO_CLI_PACKAGE_VERSION",
            help="The version of the package to publish.",
        ),
    ] = "",
    dry_run: Annotated[
        bool,
        typer.Option("--dry-run", "-n", help="Dry run the package publication."),
    ] = False,
    package_address: Annotated[
        str | None,
        typer.Argument(help="The address of the package to publish."),
    ] = None,
    skip_auth: Annotated[
        bool,
        typer.Option("--skip-auth", "-s", help="Skip authentication."),
    ] = False,
    skip_duplicate_versions: Annotated[
        bool,
        typer.Option(
            "--skip-duplicate-versions",
            envvar="ATO_CLI_PACKAGE_SKIP_DUPLICATE_VERSIONS",
            help="Ignore duplicate version errors.",
        ),
    ] = False,
):
    """
    Publish a package to the package registry.

    Currently, the only supported authentication method is Github Actions OIDC.

    For the options which allow multiple inputs, use comma separated values.
    """

    # Apply the entry-point early
    # This will configure the project root properly, meaning you can practically spec
    # the working directory of the publish and expands for future use publishing
    # packagelets from specific module entrypoints
    config.apply_options(entry=package_address)
    logger.info("Using project config: %s", config.project.paths.root / "ato.yaml")

    if version:  # NOT `is not None` to allow for empty strings
        _apply_version(version)

    if config.project.package is None:
        raise UserException(
            "Project has no package configuration. "
            "Please add a `package` section to your `ato.yaml` file."
        )
    logger.info("Package version: %s", config.project.package.version)

    if not config.project.package.identifier:
        raise UserBadParameterError(
            "Project `identifier` is not set. Set via ATO_PACKAGE_IDENTIFIER envvar"
            " or in `ato.yaml`"
        )

    if not config.project.package.repository:
        raise UserBadParameterError(
            "Project `repository` is not set. Set via ATO_PACKAGE_REPOSITORY envvar"
            " or in `ato.yaml`"
        )

    # Build the package
    try:
        dist = Dist.build_dist(
            cfg=config.project,
            output_path=config.project.paths.build,
        )
    except DistValidationError as e:
        raise UserException(
            "Could not build package distribution: %s" % e
            + "\n Have a look at https://docs.atopile.io/atopile/essentials/4-packages"
        ) from e

    logger.info("Package distribution built: %s", dist.path)

    # Upload sequence
    if dry_run:
        logger.info("Dry run, skipping upload")
    else:
        api = PackagesAPIClient()
        try:
            package_url = api.publish(
                identifier=config.project.package.identifier,
                version=str(config.project.package.version),
                dist=dist,
                skip_auth=skip_auth,
            ).url
        except packages_api.Errors.ReleaseAlreadyExistsError:
            if skip_duplicate_versions:
                logger.info("Release already exists, skipping")
                return
            else:
                raise

        logger.info("Package URL: %s", package_url)

    logger.info("Done! 📦🛳️")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class CH340x(Module):
    usb: F.USB2_0
    uart: F.UART
    tnow: F.ElectricLogic
    gpio_power: F.ElectricPower

    designator = L.f_field(F.has_designator_prefix)(F.has_designator_prefix.Prefix.U)
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://wch-ic.com/downloads/file/79.html"
    )

    def __preinit__(self):
        self.gpio_power.lv.connect(self.usb.usb_if.buspower.lv)

        self.gpio_power.voltage.constrain_subset(L.Range(0 * P.V, 5.3 * P.V))
        self.usb.usb_if.buspower.voltage.constrain_subset(L.Range(4 * P.V, 5.3 * P.V))

        self.gpio_power.lv.connect(self.usb.usb_if.buspower.lv)

    @L.rt_field
    def can_be_decoupled(self):
        class _(F.can_be_decoupled.impl()):
            def decouple(self, owner: Module):
                obj = self.get_obj(CH340x)

                obj.gpio_power.decoupled.decouple(owner=owner)
                obj.usb.usb_if.buspower.decoupled.decouple(owner=owner)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import pytest

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.app.checks import (
    RequiresExternalUsageNotFulfilled,
    check_requires_external_usage,
)

logger = logging.getLogger(__name__)


def test_requires_external_usage():
    class App(Module):
        class Outer(Module):
            class Inner(Module):
                b: F.Electrical

            a: F.Electrical
            inner: Inner

            def __preinit__(self):
                self.a.add(F.requires_external_usage())

        outer1: Outer
        outer2: Outer

    app = App()

    # no connections
    with pytest.raises(RequiresExternalUsageNotFulfilled):
        check_requires_external_usage(app, app.get_graph())

    # internal connection
    app.outer1.a.connect(app.outer1.inner.b)
    with pytest.raises(RequiresExternalUsageNotFulfilled):
        check_requires_external_usage(app, app.get_graph())

    # path to external
    app.outer1.inner.b.connect(app.outer2.a)
    with pytest.raises(RequiresExternalUsageNotFulfilled):
        check_requires_external_usage(app, app.get_graph())

    # direct external connection
    app.outer1.a.connect(app.outer2.inner.b)
    check_requires_external_usage(app, app.get_graph())



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import json
import logging
from pathlib import Path

from easyeda2kicad.easyeda.easyeda_api import EasyedaApi
from easyeda2kicad.easyeda.easyeda_importer import (
    Easyeda3dModelImporter,
    EasyedaFootprintImporter,
    EasyedaSymbolImporter,
)
from easyeda2kicad.easyeda.parameters_easyeda import EeSymbol
from easyeda2kicad.kicad.export_kicad_3d_model import Exporter3dModelKicad
from easyeda2kicad.kicad.export_kicad_footprint import ExporterFootprintKicad
from easyeda2kicad.kicad.export_kicad_symbol import ExporterSymbolKicad, KicadVersion

import faebryk.library._F as F
from atopile.config import config
from faebryk.core.module import Module
from faebryk.libs.picker.localpick import PickerOption
from faebryk.libs.picker.picker import (
    Part,
    Supplier,
)
from faebryk.libs.util import ConfigFlag

logger = logging.getLogger(__name__)

CRAWL_DATASHEET = ConfigFlag(
    "LCSC_DATASHEET", default=False, descr="Crawl for datasheet on LCSC"
)

EASYEDA_CACHE_FOLDER = Path("cache/easyeda")

EXPORT_NON_EXISTING_MODELS = False

"""
easyeda2kicad has not figured out 100% yet how to do model translations.
It's unfortunately also not really easy.
A lot of SMD components (especially passives, ICs, etc) seem to be doing just fine with
an x,y translation of 0. However that makes some other SMD components behave even worse.
Since in a typical design most components are passives etc, this workaround can save
a lot of time and manual work.
"""
WORKAROUND_SMD_3D_MODEL_FIX = True

"""
Some THT models seem to be fixed when assuming their translation is mm instead of inch.
Does not really make a lot of sense.
"""
WORKAROUND_THT_INCH_MM_SWAP_FIX = False


def _fix_3d_model_offsets(ki_footprint):
    if ki_footprint.output.model_3d is None:
        return

    if WORKAROUND_SMD_3D_MODEL_FIX:
        if ki_footprint.input.info.fp_type == "smd":
            ki_footprint.output.model_3d.translation.x = 0
            ki_footprint.output.model_3d.translation.y = 0
    if WORKAROUND_THT_INCH_MM_SWAP_FIX:
        if ki_footprint.input.info.fp_type != "smd":
            ki_footprint.output.model_3d.translation.x *= 2.54
            ki_footprint.output.model_3d.translation.y *= 2.54


class LCSCException(Exception):
    def __init__(self, partno: str, *args: object) -> None:
        self.partno = partno
        super().__init__(*args)

    def __str__(self) -> str:
        return f"{type(self).__name__}: {self.partno} - {self.args}"


class LCSC_NoDataException(LCSCException): ...


class LCSC_PinmapException(LCSCException): ...


def get_raw(lcsc_id: str):
    api = EasyedaApi()

    cache_base = config.project.paths.build / EASYEDA_CACHE_FOLDER
    cache_base.mkdir(parents=True, exist_ok=True)

    comp_path = cache_base.joinpath(lcsc_id)
    if not comp_path.exists():
        logger.debug(f"Did not find component {lcsc_id} in cache, downloading...")
        cad_data = api.get_cad_data_of_component(lcsc_id=lcsc_id)
        serialized = json.dumps(cad_data)
        comp_path.write_text(serialized, encoding="utf-8")

    data = json.loads(comp_path.read_text(encoding="utf-8"))

    # API returned no data
    if not data:
        raise LCSC_NoDataException(
            lcsc_id, f"Failed to fetch data from EasyEDA API for part {lcsc_id}"
        )

    return data


def download_easyeda_info(lcsc_id: str, get_model: bool = True):
    # easyeda api access & caching --------------------------------------------
    data = get_raw(lcsc_id)

    easyeda_footprint = EasyedaFootprintImporter(
        easyeda_cp_cad_data=data
    ).get_footprint()

    easyeda_symbol = EasyedaSymbolImporter(easyeda_cp_cad_data=data).get_symbol()

    # paths -------------------------------------------------------------------
    name = easyeda_footprint.info.name
    out_base_path = config.project.paths.component_lib
    fp_base_path = out_base_path / "footprints" / "lcsc.pretty"  # TODO: config property
    sym_base_path = out_base_path / "lcsc.kicad_sym"
    model_base_path = out_base_path / "lcsc"
    fp_base_path.mkdir(exist_ok=True, parents=True)
    footprint_filename = f"{name}.kicad_mod"
    footprint_filepath = fp_base_path.joinpath(footprint_filename)

    # The base_path has to be split from the full path, because the exporter
    # will append .3dshapes to it
    model_base_path_full = model_base_path.with_suffix(".3dshapes")
    model_base_path_full.mkdir(exist_ok=True, parents=True)

    # export to kicad ---------------------------------------------------------
    ki_footprint = ExporterFootprintKicad(easyeda_footprint)
    ki_symbol = ExporterSymbolKicad(easyeda_symbol, KicadVersion.v6)

    _fix_3d_model_offsets(ki_footprint)

    easyeda_model = Easyeda3dModelImporter(
        easyeda_cp_cad_data=data, download_raw_3d_model=False
    ).output

    ki_model = None
    if easyeda_model:
        ki_model = Exporter3dModelKicad(easyeda_model)

    if easyeda_model is not None:
        model_path = model_base_path_full / f"{easyeda_model.name}.wrl"
        if get_model and not model_path.exists():
            logger.debug(f"Downloading & Exporting 3dmodel {model_path}")
            easyeda_model = Easyeda3dModelImporter(
                easyeda_cp_cad_data=data, download_raw_3d_model=True
            ).output
            assert easyeda_model is not None
            ki_model = Exporter3dModelKicad(easyeda_model)
            ki_model.export(str(model_base_path))

        if not model_path.exists() and not EXPORT_NON_EXISTING_MODELS:
            ki_footprint.output.model_3d = None
    elif get_model:
        logger.warning(f"No 3D model for '{name}'")

    if not footprint_filepath.exists():
        logger.debug(f"Exporting footprint {footprint_filepath}")
        try:
            kicad_model_path = str(
                "${KIPRJMOD}"
                / model_base_path_full.relative_to(
                    config.build.paths.kicad_project.parent, walk_up=True
                )
            )
        except RuntimeError:
            # FIXME: this shouldn't need to exist
            # It's a workaround that'll only work for a single user.
            kicad_model_path = str(model_base_path_full.resolve())

        logger.debug(f"Exporting 3D model to: {kicad_model_path}")
        ki_footprint.export(
            footprint_full_path=str(footprint_filepath),
            model_3d_path=kicad_model_path,
        )

    if not sym_base_path.exists():
        logger.debug(f"Exporting symbol {sym_base_path}")
        ki_symbol.export(str(sym_base_path))

    return (
        ki_footprint,
        ki_model,
        easyeda_footprint,
        easyeda_model,
        easyeda_symbol,
        footprint_filepath,
    )


def get_datasheet_url(part: EeSymbol):
    # TODO use easyeda2kicad api as soon as works again
    # return part.info.datasheet

    if not CRAWL_DATASHEET:
        return None

    import re

    import requests

    url = part.info.datasheet
    if not url:
        return None
    # make requests act like curl
    lcsc_site = requests.get(url, headers={"User-Agent": "curl/7.81.0"})
    lcsc_id = part.info.lcsc_id
    # find _{partno}.pdf in html
    match = re.search(f'href="(https://[^"]+_{lcsc_id}.pdf)"', lcsc_site.text)
    if match:
        pdfurl = match.group(1)
        logger.debug(f"Found datasheet for {lcsc_id} at {pdfurl}")
        return pdfurl
    else:
        return None


def check_attachable(component: Module):
    if not component.has_trait(F.has_footprint):
        if not component.has_trait(F.can_attach_to_footprint):
            if not component.has_trait(F.has_pin_association_heuristic):
                raise LCSC_PinmapException(
                    "",
                    f"Need either F.can_attach_to_footprint or "
                    "F.has_pin_association_heuristic"
                    f" for {component}",
                )


def attach(
    component: Module, partno: str, get_model: bool = True, check_only: bool = False
):
    try:
        _, _, easyeda_footprint, _, easyeda_symbol, footprint_filepath = (
            download_easyeda_info(partno, get_model=get_model)
        )
    except LCSC_NoDataException:
        if component.has_trait(F.has_footprint):
            easyeda_symbol = None
            easyeda_footprint = None
            footprint_filepath = None
        else:
            raise

    # TODO maybe check the symbol matches, even if a footprint is already attached?
    if not component.has_trait(F.has_footprint):
        assert easyeda_symbol is not None
        assert easyeda_footprint is not None
        assert footprint_filepath is not None
        if not component.has_trait(F.can_attach_to_footprint):
            # TODO make this a trait
            pins = [
                (pin.settings.spice_pin_number, pin.name.text)
                for unit in easyeda_symbol.units
                for pin in unit.pins
            ]
            try:
                pinmap = component.get_trait(F.has_pin_association_heuristic).get_pins(
                    pins
                )
            except F.has_pin_association_heuristic.PinMatchException as e:
                raise LCSC_PinmapException(partno, f"Failed to get pinmap: {e}") from e

            if check_only:
                return

            component.add(F.can_attach_to_footprint_via_pinmap(pinmap))

            sym = F.Symbol.with_component(component, pinmap)
            sym.add(F.Symbol.has_kicad_symbol(f"lcsc:{easyeda_footprint.info.name}"))

        if check_only:
            return

        # footprint
        fp = F.KicadFootprint([p.number for p in easyeda_footprint.pads])
        fp.add(F.KicadFootprint.has_file(footprint_filepath))
        component.get_trait(F.can_attach_to_footprint).attach(fp)

    if check_only:
        return

    component.add(F.has_descriptive_properties_defined({"LCSC": partno}))

    if easyeda_symbol is not None:
        datasheet = get_datasheet_url(easyeda_symbol)
        if datasheet:
            component.add(F.has_datasheet_defined(datasheet))

    # model done by kicad (in fp)


class LCSC(Supplier):
    def attach(self, module: Module, part: PickerOption):
        assert isinstance(part.part, LCSC_Part)
        attach(component=module, partno=part.part.partno)
        if part.info is not None:
            module.add(F.has_descriptive_properties_defined(part.info))


class LCSC_Part(Part):
    def __init__(self, partno: str) -> None:
        super().__init__(partno=partno, supplier=LCSC())



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class XT30PW2plus2(Module):
    """
    XT30 connector with 2 high current and 2 low current pins.
    90 degree orientation PCB mount.
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    power: F.ElectricPower
    contact = L.list_field(2, F.Electrical)

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )
    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "Changzhou Amass Elec",
            DescriptiveProperties.partno: "XT30PW(2+2)-M.G.B",
        }
    )
    lcsc_id = L.f_field(F.has_descriptive_properties_defined)({"LCSC": "C19268030"})
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://wmsc.lcsc.com/wmsc/upload/file/pdf/v2/lcsc/2401121133_Changzhou-Amass-Elec-XT30-2-2PW-M-G-B_C19268030.pdf"
    )

    @L.rt_field
    def attach_via_pinmap(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.power.hv,
                "2": self.power.lv,
                "3": self.contact[0],
                "4": self.contact[1],
            }
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import os
import shutil
import sys
from pathlib import Path

import psutil

import faebryk.library._F as F
from atopile.cli.logging import ALERT
from atopile.config import config
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.core.node import Node, NodeException
from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
from faebryk.exporters.pcb.routing.util import apply_route_in_pcb
from faebryk.libs.exceptions import UserResourceException, downgrade
from faebryk.libs.kicad.fileformats_latest import (
    C_kicad_fp_lib_table_file,
    C_kicad_project_file,
)
from faebryk.libs.util import (
    cast_assert,
    duplicates,
    groupby,
    hash_string,
    md_list,
    not_none,
    once,
    remove_venv_from_env,
)

logger = logging.getLogger(__name__)


def apply_layouts(app: Module):
    if not app.has_trait(F.has_pcb_position):
        app.add(
            F.has_pcb_position_defined(
                F.has_pcb_position.Point((0, 0, 0, F.has_pcb_position.layer_type.NONE))
            )
        )

    for level in app.get_tree(types=Node).iter_by_depth():
        for n in level:
            if n.has_trait(F.has_pcb_layout):
                n.get_trait(F.has_pcb_layout).apply()


def apply_routing(app: Module, transformer: PCB_Transformer):
    strategies: list[tuple[F.has_pcb_routing_strategy, int]] = []

    for i, level in enumerate(app.get_tree(types=Node).iter_by_depth()):
        for n in level:
            if not n.has_trait(F.has_pcb_routing_strategy):
                continue

            strategies.append((n.get_trait(F.has_pcb_routing_strategy), i))

    logger.info("Applying routes")

    # sort by (prio, level)
    for strategy, level in sorted(
        strategies, key=lambda x: (x[0].priority, x[1]), reverse=True
    ):
        logger.debug(f"{strategy} | {level=}")

        routes = strategy.calculate(transformer)
        for route in routes:
            apply_route_in_pcb(route, transformer)


def ensure_footprint_lib(
    lib_name: str, fppath: os.PathLike, fptable: C_kicad_fp_lib_table_file | None = None
):
    fppath = Path(fppath)

    if fptable is None:
        try:
            fptable = C_kicad_fp_lib_table_file.loads(config.build.paths.fp_lib_table)
        except FileNotFoundError:
            fptable = C_kicad_fp_lib_table_file.skeleton()

    relative = True
    try:
        fppath_rel = fppath.resolve().relative_to(
            config.build.paths.fp_lib_table.parent.resolve(), walk_up=True
        )
        # check if not going up outside the project directory
        # relative_to raises a ValueError if it has to walk up to make a relative path
        fppath.relative_to(not_none(config.project.paths.root))
    except ValueError:
        relative = False
        with downgrade(UserResourceException):
            raise UserResourceException(
                f"Footprint path {fppath} is outside the project directory."
                "This is unstable behavior and may be deprecated in the future."
            )
    else:
        fppath = fppath_rel

    uri = str(fppath)

    if relative:
        assert not uri.startswith("/")
        assert not uri.startswith("${KIPRJMOD}")
        uri = "${KIPRJMOD}/" + uri

    lib = C_kicad_fp_lib_table_file.C_fp_lib_table.C_lib(
        name=lib_name,
        type="KiCad",
        uri=uri,
        options="",
        descr=f"atopile: {lib_name} footprints",
    )

    matching_libs = [
        lib_ for lib_ in fptable.fp_lib_table.libs if lib_.name == lib.name
    ]
    lib_is_duplicated = len(matching_libs) != 1
    lib_is_outdated = any(lib_ != lib for lib_ in matching_libs)

    if lib_is_duplicated or lib_is_outdated:
        fptable.fp_lib_table.libs = [
            lib for lib in fptable.fp_lib_table.libs if lib.name != lib_name
        ] + [lib]

        logger.log(ALERT, "pcbnew restart required (updated fp-lib-table)")

    fptable.dumps(config.build.paths.fp_lib_table)

    return fptable


@once
def find_pcbnew() -> os.PathLike:
    """Figure out what to call for the pcbnew CLI."""
    if sys.platform.startswith("linux"):
        return Path("pcbnew")

    if sys.platform.startswith("darwin"):
        base = Path("/Applications/KiCad/")
    elif sys.platform.startswith("win"):
        base = Path(not_none(os.getenv("ProgramFiles"))) / "KiCad"
    else:
        raise NotImplementedError(f"Unsupported platform: {sys.platform}")

    if path := list(base.glob("**/pcbnew")):
        # TODO: find the best version
        return path[0]

    raise FileNotFoundError("Could not find pcbnew executable")


def open_pcb(pcb_path: os.PathLike):
    import subprocess

    pcbnew = find_pcbnew()

    # Check if pcbnew is already running with this pcb
    for process in psutil.process_iter(["name", "cmdline"]):
        if process.info["name"] and "pcbnew" in process.info["name"].lower():
            if process.info["cmdline"] and str(pcb_path) in process.info["cmdline"]:
                raise RuntimeError(f"PCBnew is already running with {pcb_path}")

    subprocess.Popen(
        [str(pcbnew), str(pcb_path)],
        env=remove_venv_from_env(),
        stderr=subprocess.DEVNULL,
    )


def set_kicad_netlist_path_in_project(project_path: Path, netlist_path: Path):
    """
    Set netlist path in gui menu
    """
    if not project_path.exists():
        project = C_kicad_project_file()
    else:
        project = C_kicad_project_file.loads(project_path)
    project.pcbnew.last_paths.netlist = str(
        netlist_path.resolve().relative_to(project_path.parent.resolve(), walk_up=True)
    )
    project.dumps(project_path)


def load_net_names(graph: Graph, raise_duplicates: bool = True) -> None:
    """
    Load nets from attached footprints and attach them to the nodes.
    """

    net_names: dict[F.Net, str] = {
        cast_assert(F.Net, net): pcb_net_t.get_net().name
        for net, pcb_net_t in GraphFunctions(graph).nodes_with_trait(
            PCB_Transformer.has_linked_kicad_net
        )
    }

    if dups := duplicates(net_names.values(), lambda x: x):
        counts_by_net = [f"{k} (x{len(v)})" for k, v in dups.items()]
        with downgrade(UserResourceException, raise_anyway=raise_duplicates):
            # TODO: origin information
            raise UserResourceException(
                f"Multiple nets are named the same:\n{md_list(counts_by_net)}"
            )

    for net, name in net_names.items():
        net.add(F.has_overriden_name_defined(name))


def check_net_names(graph: Graph):
    """Raise an error if any nets have the same name."""
    gf = GraphFunctions(graph)
    nets = gf.nodes_of_type(F.Net)

    net_name_collisions = {
        k: v
        for k, v in groupby(
            nets, lambda n: n.get_trait(F.has_overriden_name).get_name()
        ).items()
        if len(v) > 1
    }
    if net_name_collisions:
        raise UserResourceException(f"Net name collision: {net_name_collisions}")


def create_footprint_library(app: Module, no_fp_lib: bool = False) -> None:
    """
    Ensure all KicadFootprints have a kicad identifier (via the F.has_kicad_footprint
    trait).

    Create a footprint library for all the footprints with files and without KiCAD
    identifiers.

    Check all of the KicadFootprints have a manual identifier. Raise an error if they
    don't.

    Args:
        no_fp_lib: If True, don't create a footprint library (only useful for testing)
    """
    from atopile.packages import KNOWN_PACKAGES_TO_FOOTPRINT

    package_fp_paths = set(KNOWN_PACKAGES_TO_FOOTPRINT.values())

    LIB_NAME = "atopile"

    # Create the library it doesn't exist
    atopile_fp_dir = config.project.paths.get_footprint_lib("atopile")
    atopile_fp_dir.mkdir(parents=True, exist_ok=True)
    if not no_fp_lib:
        ensure_footprint_lib(LIB_NAME, atopile_fp_dir)

    # Cache the mapping from path to identifier and the path to the new file
    path_map: dict[Path, tuple[str, Path]] = {}

    for fp in app.get_children(direct_only=False, types=F.KicadFootprint):
        # has_kicad_identifier implies has_kicad_footprint, but not the other way around
        if fp.has_trait(F.has_kicad_footprint):
            # I started writing this and forgot where I was going
            # So I'm going to leave it an attempt to prompt my or someone else's ideas
            # as to what we were supposed to do here
            pass
        else:
            if has_file_t := fp.try_get_trait(F.KicadFootprint.has_file):
                path = has_file_t.file
                # We privilege packages and assume it's what's driving
                if path in package_fp_paths:
                    if path in path_map:
                        id_, new_path = path_map[path]
                    else:
                        id_ = f"{LIB_NAME}:{path.stem}"
                        new_path = atopile_fp_dir / path.name
                        shutil.copy(path, new_path)
                        path_map[path] = (id_, new_path)

                else:
                    try:
                        prj_rel_path = path.relative_to(config.project.paths.root)

                    # Raised when the file isn't relative to the project directory
                    except ValueError as ex:
                        raise UserResourceException(
                            f"Footprint file {path} is outside the project"
                            " directory. Footprint files must be in the project"
                            " directory.",
                            markdown=False,
                        ) from ex

                    # Copy the footprint to the new library with a
                    # pseudo-guaranteed unique name
                    if prj_rel_path in path_map:
                        id_, new_path = path_map[prj_rel_path]
                    else:
                        mini_hash = hash_string(str(prj_rel_path))[:6]
                        id_ = f"{LIB_NAME}:{prj_rel_path.stem}-{mini_hash}"
                        new_path = (
                            atopile_fp_dir
                            / f"{prj_rel_path.stem}-{mini_hash}{prj_rel_path.suffix}"
                        )
                        shutil.copy(path, new_path)
                        path_map[prj_rel_path] = (id_, new_path)

                fp.add(F.KicadFootprint.has_file(new_path))  # Override with new path
                # Attach the newly minted identifier
                fp.add(F.KicadFootprint.has_kicad_identifier(id_))
            else:
                # This shouldn't happen
                # KicadFootprint should always have a file, or an identifier
                raise NodeException(
                    fp, f"{fp.__class__.__name__} has no footprint identifier or file"
                )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import atexit
import shutil
import unittest
from pathlib import Path
from tempfile import mkdtemp

import pytest

import faebryk.libs.picker.lcsc as lcsc

"""
This mode is for when you want to check the models in kicad.
This is especially useful while reverse engineering the easyeda translations.
"""
INTERACTIVE_TESTING = False


@pytest.mark.usefixtures("setup_project_config")
class TestLCSC(unittest.TestCase):
    def test_model_translations(self):
        test_parts = {
            # Zero SMD
            "C1525": (0, 0, 0),
            "C2827654": (0, 0, 0),
            "C284656": (0, 0, 0),
            "C668207": (0, 0, 0),
            "C914087": (0, 0, 0),
            "C25076": (0, 0, 0),
            "C328302": (0, 0, 0),
            "C72041": (0, 0, 0),
            "C99652": (0, 0, 0),
            "C72038": (0, 0, 0),
            "C25111": (0, 0, 0),
            "C2290": (0, 0, 0),
            # Non-zero SMD
            # "C585890": (0, -3.5, 0), # TODO enable
            # "C2828092": (0, -9.4, -0.254),  # TODO enable
            # THT
            # "C5239862": (-2.159, 0, 0), #TODO enable
            # "C225521": (-6.3, 1.3, 0),  # TODO enable
        }

        if not INTERACTIVE_TESTING:
            lcsc.BUILD_FOLDER = Path(mkdtemp())
            atexit.register(lambda: shutil.rmtree(lcsc.BUILD_FOLDER))

        lcsc.LIB_FOLDER = lcsc.BUILD_FOLDER / Path("kicad/libs")
        lcsc.MODEL_PATH = None
        lcsc.EXPORT_NON_EXISTING_MODELS = True

        for part, expected in test_parts.items():
            ki_footprint, ki_model, ee_footprint, ee_model, ee_symbol, _ = (
                lcsc.download_easyeda_info(
                    part,
                    get_model=INTERACTIVE_TESTING,
                )
            )

            translation = ki_footprint.output.model_3d.translation

            self.assertEqual(
                (translation.x, translation.y, translation.z), expected, f"{part}"
            )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import json
import logging

import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.core.parameter import Parameter
from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
from faebryk.libs.sets.sets import P_Set

NO_LCSC_DISPLAY = "No LCSC number"

logger = logging.getLogger(__name__)


def load_descriptive_properties(G: Graph):
    """
    Load descriptive properties from footprints and saved parameters.
    """
    nodes = GraphFunctions(G).nodes_with_trait(
        PCB_Transformer.has_linked_kicad_footprint
    )

    for node, trait in nodes:
        assert isinstance(node, Module)
        if isinstance(node, F.Footprint):
            continue
        if node.has_trait(F.has_part_picked):
            continue

        fp = trait.get_fp()
        props = fp.propertys

        # Load LCSC number from descriptive properties
        if (
            F.has_descriptive_properties.get_from(node, "LCSC") is None
            and (lcsc_display_prop := props.get("LCSC")) is not None
            and (lcsc_display := lcsc_display_prop.value) != NO_LCSC_DISPLAY
        ):
            # node.add(F.has_explicit_part.by_supplier(lcsc_display,supplier_id="lcsc"))
            node.add(F.has_descriptive_properties_defined({"LCSC": lcsc_display}))

        # Load saved parameters from descriptive properties
        for key, value in props.items():
            if not key.startswith("PARAM_"):
                continue

            param_name = key.removeprefix("PARAM_")
            param_value = json.loads(value.value)
            param_value = P_Set.deserialize(param_value)
            param = node[param_name]
            assert isinstance(param, Parameter)
            param.alias_is(param_value)


def save_parameters(G: Graph):
    """
    Save parameters to footprints (by copying them to descriptive properties).
    """

    nodes = GraphFunctions(G).nodes_with_trait(F.has_part_picked)

    for node, _ in nodes:
        for p in node.get_children(direct_only=True, types=Parameter):
            lit = p.try_get_literal()
            if lit is None:
                continue
            lit = P_Set.from_value(lit)
            key = f"PARAM_{p.get_name()}"
            value = json.dumps(lit.serialize())
            node.add(F.has_descriptive_properties_defined({key: value}))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

""" """

# TODO BROKEN

import logging
import re

import black

from faebryk.libs.pycodegen import sanitize_name
from faebryk.libs.util import not_none

logger = logging.getLogger(__name__)

template = """
# This file was generated by faebryk netlist exporter
{header}

\"""
{docstring}
\"""
from pathlib import Path
import logging

logger = logging.getLogger(__name__)


# function imports
from faebryk.exporters.netlist.kicad.netlist_kicad import from_faebryk_t2_netlist
from faebryk.exporters.netlist.netlist import make_t2_netlist_from_t1
from faebryk.exporters.netlist.graph import (
    make_graph_from_components,
    make_t1_netlist_from_graph,
)
{function_imports}

# library imports
#TODO
{library_imports}

from faebryk.libs.examples.buildutil import export_graph, export_netlist

def run_experiment():

    {definitions}

    {connections}

    comps = [
        {comps}
    ]

    t1_ = make_t1_netlist_from_graph(make_graph_from_components(comps))

    netlist = from_faebryk_t2_netlist(make_t2_netlist_from_t1(t1_))
    assert netlist is not None

    export_netlist(netlist)
    export_graph(t1_, show=True)

# Boilerplate -----------------------------------------------------------------
import sys


def main(argc, argv, argi):
    logging.basicConfig(level=logging.INFO)

    logger.info("Running experiment")
    run_experiment()


if __name__ == "__main__":
    import os
    import sys

    root = os.path.join(os.path.dirname(__file__), "..")
    sys.path.append(root)
    main(len(sys.argv), sys.argv, iter(sys.argv))

"""

comp_template = """
    class _{name}CLS(Component):
        def __init__(self):
            super().__init__()

            class _IFs(Component.InterfacesCls()):
                unnamed = {unnamed_ifs}
                {named_if_expr}

            self = _IFs(self)

            {trait_expr}


    {ifs}
    {name} = _{name}CLS()

"""


def dict_to_str(obj):
    return "{" + "".join(f"'{k}':{v}," for k, v in obj.items()) + "}"


def str_to_str(obj):
    return f'"{obj}"'


def from_t1_netlist(t1_netlist):
    # t1_netlist = [comps{
    #   name,
    #   real,
    #   properties,
    #   neighbors,
    #   value,
    components = t1_netlist
    project = template

    comp_names = {}

    def comp_to_faebryk(component):
        def get_comp_name(component):
            if component["name"].startswith("COMP["):
                class_name = not_none(
                    re.search(r"\[(.*):.*\]", component["name"])
                ).group(1)
            else:
                class_name = component["name"]

            class_name = sanitize_name(class_name)
            assert isinstance(class_name, str)
            if re.match(pattern="^[a-zA-Z_]+[a-zA-Z_0-9]*$", string=class_name) is None:
                assert False, class_name

            ctr = comp_names.get(class_name, 0)
            name = "{type}_{ctr}".format(type=class_name, ctr=ctr)
            comp_names[class_name] = ctr + 1

            return name

        name = get_comp_name(component)

        pinmap = {
            pin: (f"{{name}}.P{pin}", f"P{pin}")
            for pin in component["neighbors"].keys()
        }

        named_if_expr = ("\n" + "    " * 4).join(
            [f"{pin_name} = Electrical()" for _, pin_name in pinmap.values()]
        )

        # Traits --------------------------------------------------
        trait_template = "".join(
            [
                "self.add({trait_name}(",
                "    {trait_args}",
                "))",
            ]
        )

        traits = []

        def add(name, args):
            traits.append(trait_template.format(trait_name=name, trait_args=args))

        add(
            "has_defined_footprint_pinmap",
            dict_to_str({k: v[0].format(name="self") for k, v in pinmap.items()}),
        )

        if component["real"]:
            add("has_defined_type_description", str_to_str(component["value"]))

            add(
                "has_defined_footprint",
                "KicadFootprint({})".format(
                    str_to_str(component["properties"]["footprint"])
                ),
            )

        trait_expr = ("\n" + "    " * 3).join(traits)

        # ---------------------------------------------------------

        comp = comp_template.format(
            ifs="",
            unnamed_ifs="[]",
            name=name,
            named_if_expr=named_if_expr,
            trait_expr=trait_expr,
        )

        return name, (comp, pinmap, component)

    named_comps = dict(map(comp_to_faebryk, components))
    name_map = {c["name"]: cname for cname, (_, __, c) in named_comps.items()}

    connections = {}
    for cname, (ccode, pinmap, comp) in sorted(
        named_comps.items(), key=lambda x: not x[1][2]["real"]
    ):
        for pin, neighbors in comp["neighbors"].items():
            for neighbor in neighbors:
                neighborcname = name_map[neighbor["vertex"]["name"]]
                npin = neighbor["pin"]
                if ((neighborcname, npin), (cname, pin)) in connections:
                    # don't add connection if symmetric connection already exists
                    continue
                connections[((cname, pin), (neighborcname, npin))] = (
                    f"{pinmap[pin][0].format(name=cname)}.connect("
                    f"{named_comps[neighborcname][1][npin][0].format(name=neighborcname)})"
                )

    project = project.format(
        header="",
        docstring="",
        function_imports="",
        library_imports="",
        definitions="\n        ".join([c[0] for c in named_comps.values()]),
        connections="\n    ".join(connections.values()),
        comps=",\n        ".join(named_comps.keys()),
    )

    project = black.format_file_contents(project, fast=False, mode=black.FileMode())
    # project = black.format_str(project, mode=black.Mode())

    return project



import logging
from pathlib import Path

from . import pullgroup, pushgroup, reloadgroup  # noqa: F401

LOG_FILE = (Path(__file__).parent / "log.log").expanduser().absolute()
LOG_FILE.parent.mkdir(parents=True, exist_ok=True)

log = logging.getLogger(__name__)
formatter = logging.Formatter("%(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
file_handler = logging.FileHandler(str(LOG_FILE), "w", "utf-8")
file_handler.setFormatter(formatter)
log.addHandler(file_handler)
log.setLevel(logging.DEBUG)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import Any, Protocol

from faebryk.core.graph import Graph
from faebryk.core.node import Node
from faebryk.core.parameter import (
    ConstrainableExpression,
    Expression,
    Parameter,
    Predicate,
)
from faebryk.libs.sets.sets import P_Set
from faebryk.libs.util import ConfigFlag

logger = logging.getLogger(__name__)

LOG_PICK_SOLVE = ConfigFlag("LOG_PICK_SOLVE", False)


class NotDeducibleException(Exception):
    def __init__(
        self,
        predicate: ConstrainableExpression,
        not_deduced: list[ConstrainableExpression],
    ):
        self.predicate = predicate
        self.not_deduced = not_deduced

    def __str__(self):
        return f"Could not deduce predicate: {self.predicate}"


class Solver(Protocol):
    def get_any_single(
        self,
        operatable: Parameter,
        lock: bool,
        suppose_constraint: Predicate | None = None,
        minimize: Expression | None = None,
    ) -> Any:
        """
        Solve for a single value for the given expression.

        Args:
            operatable: The expression or parameter to solve.
            suppose_constraint: An optional constraint that can be added to make solving
                                easier. It is only in effect for the duration of the
                                solve call.
            minimize: An optional expression to minimize while solving.
            lock: If True, ensure the result is part of the solution set of
                              the expression.

        Returns:
            A SolveResultSingle object containing the chosen value.
        """
        ...

    def try_fulfill(
        self,
        predicate: ConstrainableExpression,
        lock: bool,
        allow_unknown: bool = False,
    ) -> bool:
        """
        Try to fulfill the predicate.

        Args:
            predicate: The predicate to fulfill.
            lock: If True, ensure the result is part of the solution set of
                  the expression.

        Returns:
            True if found definite answer, False if allow_unknown and e.g Timeout or
            no deduction possible
        Raises:
            TimeoutError if not allow_unknown and timeout
            Contradiction if predicate proved false
        """
        ...

    def inspect_get_known_supersets(self, value: Parameter) -> P_Set: ...

    def update_superset_cache(self, *nodes: Node): ...

    def simplify(self, *gs: Graph | Node): ...



"""
Datatypes used in the model.
"""

import logging
from contextlib import contextmanager
from typing import Iterable, Iterator, Self

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def is_int(name: str | int | None) -> bool:
    if name is None:
        return False
    if isinstance(name, int):
        return True
    try:
        int(name)
    except ValueError:
        return False
    return True


class TypeRef(tuple[str]):
    """Shell class to provide basic utils for a reference."""

    def add_name(self, name: str | int) -> "TypeRef":
        """Return a new Ref with the given name."""
        return TypeRef((*self, str(name)))

    def __str__(self) -> str:
        return ".".join(map(str, self))

    @classmethod
    def empty(cls) -> "TypeRef":
        """Return an empty Ref."""
        return cls(())

    @classmethod
    def from_one(cls, name: str | int) -> "TypeRef":
        """Return a Ref with a single item."""
        return cls((str(name),))

    @classmethod
    def from_path_str(cls, path: str) -> "TypeRef":
        """Return a Ref from a path string."""
        return cls(path.split("."))


KeyType = str | int


class ReferencePartType:
    def __init__(self, name: str, key: KeyType | None = None):
        # TODO remove
        assert isinstance(name, str)
        self.name = name
        self.key = key

    def __str__(self) -> str:
        if self.key is not None:
            return f"{self.name}[{self.key}]"
        return self.name

    def has_key(self) -> bool:
        return self.key is not None


class FieldRef:
    """
    A class representing a field reference.
    e.g app.modules[multiresistor].resistors[0].unnamed[1]
    """

    def __init__(
        self, parts: Iterable[ReferencePartType], pin: int | str | None = None
    ):
        self.parts = list(parts)
        # shim A.1
        if pin is not None:
            ref = None
            if isinstance(pin, str) and not is_int(pin):
                ref = ReferencePartType(pin)
            else:
                # TODO: consider shiming A.0 as A.pins[0] instead of A._0
                ref = ReferencePartType(f"_{pin}")
            self.parts.append(ref)

    def __iter__(self) -> Iterator[ReferencePartType]:
        return iter(self.parts)

    def __len__(self) -> int:
        return len(self.parts)

    def __getitem__(self, index: int) -> ReferencePartType:
        return self.parts[index]

    def __contains__(self, item: ReferencePartType) -> bool:
        return item in self.parts

    def __bool__(self) -> bool:
        return bool(self.parts)

    @property
    def stem(self) -> Self:
        return type(self)(self.parts[:-1])

    @property
    def last(self) -> ReferencePartType:
        return self.parts[-1]

    def append(self, part: ReferencePartType) -> Self:
        return type(self)(self.parts + [part])

    def __str__(self) -> str:
        return ".".join(str(part) for part in self.parts)

    def to_type_ref(self) -> TypeRef | None:
        if any(part.has_key() for part in self.parts):
            return None
        return TypeRef(map(lambda x: x.name, self.parts))

    @classmethod
    def from_type_ref(cls, type_ref: TypeRef) -> "FieldRef":
        """Return a FieldRef from a TypeRef."""
        return cls(map(ReferencePartType, type_ref))


class KeyOptItem[V](tuple[TypeRef | None, V]):
    """A class representing an optionally-named thing."""

    @property
    def ref(self) -> TypeRef | None:
        """Return the name of this item, if it has one."""
        return self[0]

    @property
    def value(self) -> V:
        """Return the value of this item."""
        return self[1]

    @classmethod
    def from_kv(cls, key: TypeRef | None, value: V) -> "KeyOptItem[V]":
        """Return a KeyOptItem with a single item."""
        return KeyOptItem((key, value))


class KeyOptMap[V](tuple[KeyOptItem[V]]):
    """A class representing a set of optionally-named things."""

    def keys(self) -> Iterable[TypeRef | None]:
        """Return an iterable of all the names in this set."""
        return map(lambda x: x.ref, filter(lambda x: x.ref is not None, self))

    def values(self) -> Iterable[V]:
        """Return an iterable of all the values in this set."""
        return map(lambda x: x.value, self)

    @classmethod
    def from_item(cls, item: KeyOptItem[V]) -> "KeyOptMap[V]":
        """Return a KeyOptMap with a single item."""
        return KeyOptMap((item,))

    @classmethod
    def from_kv(cls, key: TypeRef | None, value: V) -> "KeyOptMap[V]":
        """Return a KeyOptMap with a single item."""
        return cls.from_item(KeyOptItem.from_kv(key, value))

    @classmethod
    def empty(cls) -> "KeyOptMap[V]":
        """Return an empty KeyOptMap."""
        return cls(())


class StackList[V](list[V]):
    """Manages context while compiling ato code."""

    def __init__(self) -> None:
        super().__init__([])

    @contextmanager
    def enter(self, context: V):
        """Enter a context."""
        self.append(context)
        try:
            yield
        finally:
            self.pop()

    @property
    def top(self) -> V:
        """Return the current context."""
        return self[-1]



import logging
from os import PathLike
from pathlib import Path

from antlr4 import CommonTokenStream, FileStream, InputStream, Token
from antlr4.error.ErrorListener import ErrorListener

from atopile.parser.AtoLexer import AtoLexer
from atopile.parser.AtoParser import AtoParser

from .errors import UserFileNotFoundError, UserSyntaxError

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class ErrorListenerConverter(ErrorListener):
    """Converts an error into an AtoSyntaxError."""

    def syntaxError(
        self,
        recognizer,
        offendingSymbol: Token,
        line: int,
        column: int,
        msg: str,
        e: Exception | None,
    ):
        if e is None:
            msg = msg
        else:
            msg = f"{str(e)} '{msg}'"

        input_stream: CommonTokenStream = recognizer.getInputStream()
        # This fill is required to get context past the offending symbol
        # to accurately report the error
        input_stream.fill()

        raise UserSyntaxError.from_tokens(
            input_stream, offendingSymbol, None, msg, markdown=False
        )


def make_parser(src_stream: InputStream) -> AtoParser:
    """Make a parser from a stream."""
    lexer = AtoLexer(src_stream)
    stream = CommonTokenStream(lexer)
    parser = AtoParser(stream)

    parser.removeErrorListeners()
    parser.addErrorListener(ErrorListenerConverter())

    return parser


def parse_text_as_file(
    src_code: str, src_path: None | str | Path = None
) -> AtoParser.File_inputContext:
    """Parse a string as a file input."""
    input = InputStream(src_code)
    input.name = src_path
    parser = make_parser(input)

    tree = parser.file_input()

    return tree


def parse_file(src_path: Path) -> AtoParser.File_inputContext:
    """Parse a file from a path."""
    input = FileStream(str(src_path), encoding="utf-8")
    input.name = src_path
    parser = make_parser(input)

    tree = parser.file_input()

    return tree


class FileParser:
    """Parses a file."""

    def __init__(self) -> None:
        self.cache = {}

    def get_ast_from_file(self, src_origin: PathLike) -> AtoParser.File_inputContext:
        """Get the AST from a file."""

        src_origin_str = str(src_origin)
        src_origin_path = Path(src_origin)

        if src_origin_str not in self.cache:
            if not src_origin_path.exists():
                raise UserFileNotFoundError(src_origin_str)
            self.cache[src_origin_str] = parse_file(src_origin_path)

        return self.cache[src_origin_str]


parser = FileParser()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path

from faebryk.core.module import Module
from faebryk.exporters.bom.jlcpcb import write_bom_jlcpcb
from faebryk.exporters.pcb.kicad.artifacts import (
    export_dxf,
    export_gerber,
    export_glb,
    export_pick_and_place,
    export_step,
)
from faebryk.exporters.pcb.pick_and_place.jlcpcb import (
    convert_kicad_pick_and_place_to_jlcpcb,
)

logger = logging.getLogger(__name__)


def export_pcba_artifacts(out: Path, pcb_path: Path, app: Module):
    cad_path = out.joinpath("cad")
    cad_path.mkdir(parents=True, exist_ok=True)

    logger.info("Exporting PCBA artifacts")

    write_bom_jlcpcb(
        app.get_children_modules(types=Module), out.joinpath("jlcpcb_bom.csv")
    )
    export_step(pcb_path, step_file=cad_path.joinpath("pcba.step"))
    export_glb(pcb_path, glb_file=cad_path.joinpath("pcba.glb"))
    export_dxf(pcb_path, dxf_file=cad_path.joinpath("pcba.dxf"))
    export_gerber(pcb_path, gerber_zip_file=out.joinpath("gerber.zip"))
    pnp_file = out.joinpath("pick_and_place.csv")
    export_pick_and_place(pcb_path, pick_and_place_file=pnp_file)
    convert_kicad_pick_and_place_to_jlcpcb(
        pnp_file,
        out.joinpath("jlcpcb_pick_and_place.csv"),
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import pytest
from pint import Quantity

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.sets.quantity_sets import Quantity_Interval
from faebryk.libs.units import P
from faebryk.libs.util import cast_assert


@pytest.mark.parametrize(
    "capacitances",
    [
        [(1, 100 * P.uF, 0.2)],
        [(2, 100 * P.uF, 0.2)],
        [(1, 100 * P.uF, 0.2), (1, 200 * P.nF, 0.2)],
        [(2, 100 * P.uF, 0.2), (2, 200 * P.nF, 0.2)],
        [(3, 100 * P.uF, 0.2), (2, 200 * P.nF, 0.2), (2, 300 * P.nF, 0.2)],
    ],
)
def test_decouple(capacitances: list[tuple[int, Quantity, float]]):
    class App(Module):
        power: F.ElectricPower

    app = App()

    total_count = 0
    for count, capacitance, tolerance in capacitances:
        total_count += count
        app.power.decoupled.decouple(app, count=count).explicit(capacitance, tolerance)

    cap = app.power.get_trait(F.is_decoupled).capacitor

    caps = cap.get_children_modules(
        types=F.Capacitor, f_filter=lambda c: type(c) is F.Capacitor, include_root=True
    )

    expected = sorted(
        [
            L.Range.from_center_rel(capacitance, tolerance)
            for count, capacitance, tolerance in capacitances
            for _ in range(count)
        ],
        key=lambda x: x.min_elem,
    )

    # TODO remove, this is the alternative to the solver
    lits = [c.capacitance.try_get_literal_subset() for c in caps]
    lits = sorted(lits, key=lambda x: cast_assert(Quantity_Interval, x).min_elem)

    assert lits == expected



from typing import Callable, Sequence

from faebryk.core.node import Node
from faebryk.core.trait import Trait


class implements_design_check(Trait.TraitT.decless()):
    class UnfulfilledCheckException(Exception):
        nodes: Sequence[Node]

        def __init__(self, message: str, nodes: Sequence[Node]):
            self.nodes = nodes
            super().__init__(message)

    class MaybeUnfulfilledCheckException(UnfulfilledCheckException): ...

    def __init__(self, check: Callable[[], None]):
        super().__init__()
        self.check = check



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_equal_pins_in_ifs(F.has_equal_pins.impl()):
    def get_pin_map(self):
        return {
            p: str(i + 1)
            for i, p in enumerate(self.obj.get_children(direct_only=True, types=F.Pad))
        }



import io
import logging
import sys
import warnings
from collections import defaultdict
from dataclasses import dataclass
from datetime import timedelta
from functools import partial, reduce
from math import inf
from operator import add, mul, pow, sub, truediv
from typing import Any, Callable, Iterable

import pytest  # noqa: F401
import typer
from hypothesis import HealthCheck, Phase, example, given, settings
from hypothesis import strategies as st
from hypothesis.core import reproduce_failure  # noqa: F401
from hypothesis.errors import NonInteractiveExampleWarning
from hypothesis.strategies._internal.lazy import LazyStrategy
from rich.console import Console
from rich.table import Table

from faebryk.core.core import Namespace
from faebryk.core.graph import GraphFunctions
from faebryk.core.parameter import (
    Abs,
    Add,
    Arithmetic,
    Ceil,
    Cos,
    Divide,
    Floor,
    Functional,
    Log,
    Max,
    Min,
    Multiply,
    Numbers,
    Parameter,
    Power,
    Round,
    Sin,
    Sqrt,
    Subtract,
)
from faebryk.core.solver.defaultsolver import DefaultSolver
from faebryk.core.solver.utils import Contradiction, get_graphs
from faebryk.libs.library.L import Range
from faebryk.libs.sets.numeric_sets import float_round
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval,
    Quantity_Interval_Disjoint,
)
from faebryk.libs.test.times import Times
from faebryk.libs.units import Quantity
from faebryk.libs.util import ConfigFlag, ConfigFlagInt, groupby, once

logger = logging.getLogger(__name__)

# Workaround: for large repr generation of hypothesis strategies,
LazyStrategy.__repr__ = lambda self: str(id(self))

NUM_EXAMPLES = ConfigFlagInt(
    "ST_NUMEXAMPLES",
    default=100,
    descr="Number of examples to run for fuzzer",
)
ENABLE_PROGRESS_TRACKING = ConfigFlag("ST_PROGRESS", default=False)
CATCH_EVALUATION_ERRORS = True

ABS_UPPER_LIMIT = 1e12  # Terra or inf
DIGITS_LOWER_LIMIT = 12  # pico
ABS_LOWER_LIMIT = 10**-DIGITS_LOWER_LIMIT
ALLOW_ROUNDING_ERROR = True
ALLOW_EVAL_ERROR = True


operator_map: dict[type[Arithmetic], Callable] = {
    Add: add,
    Subtract: sub,
    Multiply: mul,
    Divide: truediv,
    Sqrt: lambda x: x.op_sqrt(),
    Power: pow,
    Round: float_round,
    Abs: abs,
    Sin: lambda x: x.op_sin(),
    Log: lambda x: x.op_log(),
    Cos: lambda x: x.op_cos(),
    Floor: lambda x: x.op_floor(),
    Ceil: lambda x: x.op_ceil(),
    Min: min,
    Max: max,
}


def lit(val) -> Quantity_Interval_Disjoint:
    return Quantity_Interval_Disjoint.from_value(val)


def p(val):
    return Builders.build_parameter(lit(val))


class Builders(Namespace):
    @staticmethod
    def build_parameter(quantity) -> Parameter:
        p = Parameter(domain=Numbers(negative=True, zero_allowed=True, integer=False))
        p.alias_is(quantity)
        return p

    @staticmethod
    def operator(op: type[Arithmetic]) -> Callable[[Iterable[Any] | Any], Arithmetic]:
        def f(operands: Iterable[Any] | Any) -> Arithmetic:
            return op(*operands) if isinstance(operands, Iterable) else op(operands)

        # required for good falsifying examples from hypothesis
        f.__name__ = f"'{op.__name__}'"

        return f


ValueT = Quantity_Interval_Disjoint | Parameter | Arithmetic


class Filters(Namespace):
    @staticmethod
    def _decorator[
        T: Callable[[ValueT], bool]
        | Callable[[ValueT | Iterable[ValueT]], bool]
        | Callable[[tuple[ValueT, ValueT]], bool]
    ](
        func: T,
    ) -> T:
        if not CATCH_EVALUATION_ERRORS:
            return func

        def _wrap(*args, **kwargs) -> bool:
            try:
                return func(*args, **kwargs)
            except Filters._EvaluationError:
                return False

        return _wrap  # type: ignore

    class _EvaluationError(Exception):
        pass

    @staticmethod
    def _unwrap_param(value: ValueT) -> Quantity_Interval_Disjoint:
        # TODO where is this coming from?
        if isinstance(value, Range):
            return lit(value)
        assert isinstance(value, ValueT)
        if isinstance(value, Parameter):
            return value.get_literal()
        elif isinstance(value, Arithmetic):
            try:
                return evaluate_expr(value)
            except Exception as e:
                raise Filters._EvaluationError(e) from e
        else:
            return value

    @_decorator
    @staticmethod
    def is_negative(value: ValueT) -> bool:
        lit = Filters._unwrap_param(value)
        return lit.max_elem < 0

    @_decorator
    @staticmethod
    def is_positive(value: ValueT) -> bool:
        lit = Filters._unwrap_param(value)
        return lit.min_elem > 0

    @_decorator
    @staticmethod
    def is_fractional(value: ValueT) -> bool:
        lit = Filters._unwrap_param(value)
        return not lit.is_integer

    @_decorator
    @staticmethod
    def is_zero(value: ValueT) -> bool:
        # no need for fancy isclose, already covered by impl
        lit = Filters._unwrap_param(value)
        return lit == 0

    @_decorator
    @staticmethod
    def crosses_zero(value: ValueT) -> bool:
        lit = Filters._unwrap_param(value)
        return 0 in lit or Filters.is_zero(lit)

    @_decorator
    @staticmethod
    def does_not_cross_zero(value: ValueT) -> bool:
        return not Filters.crosses_zero(value)

    @_decorator
    @staticmethod
    def not_empty(value: ValueT) -> bool:
        lit = Filters._unwrap_param(value)
        return not lit.is_empty()

    @_decorator
    @staticmethod
    def within_limits(value: ValueT) -> bool:
        lit = Filters._unwrap_param(value)
        abs_lit = abs(lit)
        return bool(
            (abs_lit.max_elem <= ABS_UPPER_LIMIT or abs_lit.max_elem == inf)
            and abs_lit.min_elem >= ABS_LOWER_LIMIT
        )

    @_decorator
    @staticmethod
    def all_within_limits(values: Iterable[ValueT] | ValueT) -> bool:
        if isinstance(values, ValueT):
            values = [values]
        return all(Filters.within_limits(v) for v in values)

    @staticmethod
    def no_op_overflow(op: Callable):
        @Filters._decorator
        def f(values: Iterable[ValueT] | ValueT) -> bool:
            if isinstance(values, ValueT):
                values = [values]
            lits = [Filters._unwrap_param(v) for v in values]
            try:
                expr = op(*lits)
            except Exception as e:
                raise Filters._EvaluationError(e) from e
            return Filters.within_limits(expr)

        return f

    @_decorator
    @staticmethod
    def is_valid_for_power(
        pair: tuple[ValueT, ValueT],
    ) -> bool:
        base, exp = pair
        return (
            # nan/undefined
            not (Filters.crosses_zero(base) and not Filters.is_positive(exp))
            # TODO: complex
            and not (Filters.is_fractional(exp) and not Filters.is_positive(base))
            # TODO: not impl yet
            and not Filters.crosses_zero(exp)
        )


class st_values(Namespace):
    @staticmethod
    def _numbers_with_limit(upper_limit: float, lower_limit: float):
        assert 0 <= lower_limit < 1
        ints = st.integers(
            min_value=int(-upper_limit),
            max_value=int(upper_limit),
        )

        def _floats(min_value: float, max_value: float):
            return st.decimals(
                allow_nan=False,
                allow_infinity=False,
                min_value=min_value,
                max_value=max_value,
                places=DIGITS_LOWER_LIMIT,
            )

        if lower_limit == 0:
            floats = _floats(-upper_limit, upper_limit)
        else:
            floats = st.one_of(
                _floats(lower_limit, upper_limit),
                _floats(-upper_limit, -lower_limit),
            )

        return st.one_of(ints, floats)

    numeric = _numbers_with_limit(ABS_UPPER_LIMIT, ABS_LOWER_LIMIT)

    small_numeric = _numbers_with_limit(1e2, 1e-1)

    ranges = st.builds(
        lambda values: lit(Range(*sorted(values))),
        st.tuples(
            st.one_of(st.just(-inf), numeric),
            st.one_of(st.just(inf), numeric),
        ),
    )

    small_ranges = st.builds(
        lambda values: lit(Range(*sorted(values))),
        st.tuples(small_numeric, small_numeric),
    )

    quantities = st.one_of(numeric, ranges).map(lit)

    small_quantities = st.one_of(small_numeric, small_ranges).map(lit)

    parameters = st.builds(Builders.build_parameter, quantities)

    small_parameters = st.builds(Builders.build_parameter, small_quantities)

    values = st.one_of(quantities, parameters)

    positive_values = values.filter(Filters.is_positive)

    small_values = st.one_of(small_quantities, small_parameters)

    _short_lists = partial(st.lists, min_size=1, max_size=5)
    lists = _short_lists(values)

    pairs = st.tuples(values, values)

    division_pairs = st.tuples(values, values.filter(Filters.does_not_cross_zero))

    power_pairs = st.tuples(values, small_values).filter(Filters.is_valid_for_power)


class Extension(Namespace):
    @staticmethod
    def tuples(children: st.SearchStrategy[Any]) -> st.SearchStrategy[Any]:
        return st.tuples(children, children)

    @staticmethod
    def single(children: st.SearchStrategy[Any]) -> st.SearchStrategy[Any]:
        return children

    @staticmethod
    def tuples_power(children: st.SearchStrategy[Any]) -> st.SearchStrategy[Any]:
        return st.tuples(children, st_values.small_values).filter(
            Filters.is_valid_for_power
        )

    @staticmethod
    def tuples_division(children: st.SearchStrategy[Any]) -> st.SearchStrategy[Any]:
        # TODO: exprs on the right side
        return st.tuples(
            children,
            st_values.values.filter(Filters.does_not_cross_zero),
        )

    @staticmethod
    def single_positive(
        children: st.SearchStrategy[Any],
    ) -> st.SearchStrategy[Any]:
        return children.filter(Filters.is_positive)


@dataclass
class ExprType:
    op: type[Arithmetic]
    _strategy: st.SearchStrategy[Any]
    _extension_strategy: Callable[[st.SearchStrategy[Any]], st.SearchStrategy[Any]]
    check_overflow: bool = True
    disable: bool = False

    @property
    @once
    def builder(self) -> Callable[[Any], Arithmetic]:
        return Builders.operator(self.op)

    @property
    def operator(self) -> Callable[[Any, Any], Any]:
        return operator_map[self.op]

    @property
    def strategy(self) -> st.SearchStrategy[Any]:
        out = self._strategy
        if self.check_overflow:
            # TODO why would this be needed?
            out = out.filter(Filters.all_within_limits)
            out = out.filter(Filters.no_op_overflow(self.operator))
        return out

    def extension_strategy(
        self, children: st.SearchStrategy[Any]
    ) -> st.SearchStrategy[Any]:
        out = self._extension_strategy(children)
        if self.check_overflow:
            out = out.filter(Filters.no_op_overflow(self.operator))
        return out


EXPR_TYPES = [
    ExprType(Add, st_values.pairs, Extension.tuples, check_overflow=False),
    ExprType(Subtract, st_values.pairs, Extension.tuples, check_overflow=False),
    ExprType(Multiply, st_values.pairs, Extension.tuples),
    ExprType(Divide, st_values.division_pairs, Extension.tuples_division),
    ExprType(Sqrt, st_values.positive_values, Extension.single_positive),
    ExprType(Power, st_values.power_pairs, Extension.tuples_power),
    ExprType(Log, st_values.positive_values, Extension.single_positive),
    ExprType(Sin, st_values.values, Extension.single, check_overflow=False),
    ExprType(Cos, st_values.values, Extension.single, check_overflow=False),
    ExprType(Abs, st_values.values, Extension.single, check_overflow=False),
    ExprType(Round, st_values.values, Extension.single, check_overflow=False),
    ExprType(Floor, st_values.values, Extension.single, check_overflow=False),
    ExprType(Ceil, st_values.values, Extension.single, check_overflow=False),
    # TODO
    ExprType(Min, st_values.lists, Extension.tuples, disable=True),
    ExprType(Max, st_values.lists, Extension.tuples, disable=True),
]


def test_no_forgotten_expr_types():
    ops = {expr_type.op for expr_type in EXPR_TYPES}
    import faebryk.core.parameter as P

    all_arithmetic_ops = {
        v
        for k, v in vars(P).items()
        if isinstance(v, type)
        and issubclass(v, Arithmetic)
        and not issubclass(v, Functional)
        and getattr(v, "__is_abstract__", False) is not v
    }
    assert ops == set(all_arithmetic_ops)


class st_exprs(Namespace):
    flat = st.one_of(
        *[
            st.builds(expr_type.builder, expr_type.strategy)
            for expr_type in EXPR_TYPES
            if not expr_type.disable
        ]
    )

    @staticmethod
    def _extend_tree(children: st.SearchStrategy[Any]) -> st.SearchStrategy[Any]:
        return st.one_of(
            *[
                st.builds(expr_type.builder, expr_type.extension_strategy(children))
                for expr_type in EXPR_TYPES
                if not expr_type.disable
            ]
        )

    # flat
    # op1(flat, flat) | flat
    # op2(op1 | flat, op1 | flat)
    trees = st.recursive(base=flat, extend=_extend_tree, max_leaves=20)


def evaluate_expr(
    expr: Arithmetic | Quantity,
) -> Quantity_Interval_Disjoint:
    match expr:
        # monoids
        case Add() | Multiply() | Min() | Max():
            operands = (evaluate_expr(operand) for operand in expr.operands)
            operator = operator_map[type(expr)]
            return reduce(operator, operands)
        # left/right-associative
        case Subtract() | Divide() | Power():
            operands = [evaluate_expr(operand) for operand in expr.operands]
            operator = operator_map[type(expr)]
            assert len(operands) == 2
            return operator(operands[0], operands[1])
        # unary
        case Sqrt() | Round() | Abs() | Sin() | Log() | Cos() | Floor() | Ceil():
            assert len(expr.operands) == 1
            operand = evaluate_expr(expr.operands[0])
            operator = operator_map[type(expr)]
            return operator(operand)
        case Quantity_Interval():
            # TODO: why are we getting these?
            return lit(expr)
        case Quantity_Interval_Disjoint():
            return expr
        case Parameter():
            return expr.get_literal()
        case _:
            raise ValueError(f"Unknown expression type: {type(expr)}")


@given(st_exprs.trees)
@settings(deadline=timedelta(milliseconds=1000))
def test_can_evaluate_literals(expr: Arithmetic):
    result = evaluate_expr(expr)
    assert isinstance(result, Quantity_Interval_Disjoint)


def _track():
    if not bool(ENABLE_PROGRESS_TRACKING):
        return
    if not hasattr(_track, "count"):
        _track.count = 0
    _track.count += 1
    if _track.count % 100 == 0:
        print(f"track: {_track.count}")
    return _track.count


@given(st_exprs.trees)
@settings(
    deadline=None,  # timedelta(milliseconds=1000),
    max_examples=int(NUM_EXAMPLES),
    report_multiple_bugs=False,
    phases=(
        Phase.generate,
        Phase.reuse,
        Phase.explicit,
        Phase.target,
        Phase.shrink,
        # Phase.explain,
    ),
    suppress_health_check=[
        HealthCheck.data_too_large,
        HealthCheck.too_slow,
        HealthCheck.filter_too_much,
        HealthCheck.large_base_example,
    ],
    print_blob=True,
)
def test_discover_literal_folding(expr: Arithmetic):
    """
    Disble xfail and
    run with:
    ```bash
    FBRK_ST_NUMEXAMPLES=10000 \
    FBRK_STIMEOUT=1 \
    FBRK_SMAX_ITERATIONS=10 \
    FBRK_SPARTIAL=n \
    FBRK_ST_PROGRESS=y \
    ./test/runpytest.sh -k "test_discover_literal_folding"
    ```
    """
    _track()
    solver = DefaultSolver()

    root = Parameter(domain=Numbers(negative=True, zero_allowed=True, integer=False))
    root.alias_is(expr)

    try:
        evaluated_expr = evaluate_expr(expr)
    except Exception:
        if ALLOW_EVAL_ERROR:
            return
        raise

    solver.update_superset_cache(root)
    solver_result = solver.inspect_get_known_supersets(root)

    assert isinstance(evaluated_expr, Quantity_Interval_Disjoint)
    assert isinstance(solver_result, Quantity_Interval_Disjoint)

    if ALLOW_ROUNDING_ERROR and solver_result != evaluated_expr:
        try:
            deviation_rel = evaluated_expr.op_deviation_to(solver_result, relative=True)
            return
        except Exception:
            pass
        else:
            assert deviation_rel < 0.01, f"Mismatch {evaluated_expr} != {solver_result}"

        deviation = evaluated_expr.op_deviation_to(solver_result)
        assert deviation <= 1, f"Mismatch {solver_result} != {evaluated_expr}"
    else:
        assert solver_result == evaluated_expr


# Examples -----------------------------------------------------------------------------
# --------------------------------------------------------------------------------------
@given(st_exprs.trees)
@settings(
    deadline=None,  # timedelta(milliseconds=1000),
    report_multiple_bugs=False,
    phases=(
        # Phase.reuse,
        Phase.explicit,
        Phase.target,
        # Phase.shrink,
        # Phase.explain,
    ),
    suppress_health_check=[
        HealthCheck.data_too_large,
        HealthCheck.too_slow,
        HealthCheck.filter_too_much,
        HealthCheck.large_base_example,
    ],
    print_blob=False,
)
def debug_fix_literal_folding(expr: Arithmetic):
    """
    Run with:
    ```bash
    FBRK_SPARTIAL=n
    FBRK_LOG_PICK_SOLVE=y
    FBRK_SLOG=y
    FBRK_SMAX_ITERATIONS=10
    FBRK_LOG_FMT=y
    python ./test/runtest.py -k "debug_fix_literal_folding"
    ```
    """
    solver = DefaultSolver()

    root = Parameter(domain=Numbers(negative=True, zero_allowed=True, integer=False))
    root.alias_is(expr)

    logger.info(f"expr: {expr.compact_repr()}")
    evaluated_expr = evaluate_expr(expr)
    logger.info(f"evaluated_expr: {evaluated_expr}")

    solver.update_superset_cache(root)
    solver_result = solver.inspect_get_known_supersets(root)

    assert isinstance(evaluated_expr, Quantity_Interval_Disjoint)
    correct = solver_result == evaluated_expr

    if not correct:
        logger.error(f"Failing expression: {expr.compact_repr()}")
        logger.error(f"Solver {solver_result} != {evaluated_expr} Literal")
        input()
        return
    logger.warning("PASSES")
    input()


@example(Log(Sin(Add(lit(0), lit(1)))))
@example(
    Add(
        Add(lit(0), lit(0)),
        Sqrt(Cos(lit(0))),
    ),
)
@example(
    Divide(
        Divide(
            Add(Sqrt(lit(2.0)), Subtract(lit(0), lit(0))),
            lit(891895568.0),
        ),
        lit(2.0),
    ),
)
@example(
    Divide(
        Divide(
            Sqrt(Add(lit(2))),
            lit(2),
        ),
        lit(710038921),
    )
)
@example(
    Sin(
        Sin(
            Subtract(
                lit(1),
                p(Range(-inf, inf)),
            ),
        ),
    ),
)
@example(
    Divide(
        Add(
            Subtract(lit(0), lit(1)),
            Abs(lit(Range(-inf, inf))),
        ),
        lit(1),
    ),
)
@example(
    Add(
        lit(1),
        Abs(
            Add(p(Range(-inf, inf)), p(Range(-inf, inf))),
        ),
    ),
)
@example(
    Add(
        Sqrt(lit(1)),
        Abs(lit(Range(-inf, inf))),
    ),
)
@example(
    expr=Add(
        Add(lit(-999_999_950_000)),
        Subtract(lit(50000), lit(-999_997_650_001)),
    ),
)
@example(
    expr=Subtract(
        Add(
            Add(lit(0)),
            Subtract(
                Add(lit(0)),
                Add(lit(1)),
            ),
        ),
        Add(lit(1)),
    )
)
@example(
    expr=Subtract(
        Subtract(
            lit(1),
            lit(Range(-inf, inf)),
        ),
        Add(lit(0)),
    ),
)
@example(
    Subtract(
        Abs(p(Range(-inf, inf))),
        Abs(p(Range(-inf, inf))),
    )
)
@example(Subtract(Abs(p(Range(5, 6))), Abs(p(Range(5, 6)))))
@example(
    expr=Multiply(
        Sqrt(Sqrt(lit(2))),
        Sqrt(Sqrt(lit(2))),
    )
)
@example(
    expr=Subtract(
        Round(lit(Range(2, 10))),
        Round(lit(Range(2, 10))),
    ),
)
@example(
    expr=Subtract(
        Round(
            Subtract(
                lit(0),
                lit(Range(-999_999_999_905, -0.3333333333333333)),
            ),
        ),
        Subtract(
            lit(-999_999_983_213),
            p(Range(-17297878, 999_999_992_070)),
        ),
    ),
)
@example(Abs(Round(lit(Range(-inf, inf)))))
@example(
    expr=Divide(
        Divide(
            lit(0),
            lit(Range(-1, -2.2250738585072014e-308)),
        ),
        lit(Range(-1, -2.2250738585072014e-308)),
    ),
)
@example(Multiply(Add(lit(0)), Abs(lit(Range(-inf, inf)))))
@example(Add(Add(lit(0)), Abs(p(-1))))
@example(Abs(p(-1)))
@example(expr=Round(Add(Abs(lit(0)), Round(lit(-1)))))
@example(
    expr=Add(
        Add(lit(0)),
        Multiply(Add(lit(0)), Add(lit(0))),
    )
)
@example(expr=Subtract(lit(1), lit(0))).via("discovered failure")
# --------------------------------------------------------------------------------------
@given(st_exprs.trees)
@settings(
    deadline=None,  # timedelta(milliseconds=1000),
    max_examples=10000,
    report_multiple_bugs=False,
    phases=(
        # Phase.reuse,
        Phase.explicit,
        Phase.target,
        Phase.shrink,
        Phase.explain,
    ),
    suppress_health_check=[
        HealthCheck.data_too_large,
        HealthCheck.too_slow,
        HealthCheck.filter_too_much,
        HealthCheck.large_base_example,
    ],
    print_blob=False,
)
def test_regression_literal_folding(expr: Arithmetic):
    solver = DefaultSolver()

    root = Parameter(domain=Numbers(negative=True, zero_allowed=True, integer=False))
    root.alias_is(expr)

    evaluated_expr = evaluate_expr(expr)

    solver.update_superset_cache(root)
    solver_result = solver.inspect_get_known_supersets(root)

    assert isinstance(evaluated_expr, Quantity_Interval_Disjoint)
    assert solver_result == evaluated_expr


class Stats:
    singleton: "Stats | None" = None

    @classmethod
    def get(cls) -> "Stats":
        if Stats.singleton is None:
            Stats.singleton = cls()
        return Stats.singleton

    def __init__(self):
        self.exprs: dict[Arithmetic, set[str]] = defaultdict(set)
        self.events: dict[str, set[Arithmetic]] = defaultdict(set)
        self.times = Times(multi_sample_strategy=Times.MultiSampleStrategy.ALL)
        self._total = self.times.context("total")
        self._total.__enter__()

    def print(self):
        table = Table(title="Statistics")
        table.add_column("Event", justify="left")
        table.add_column("Count", justify="right")
        table.add_column("%", justify="right")
        total = len(self.exprs)
        for name, exprs in self.events.items():
            count = len(exprs)
            percent = count / total
            table.add_row(name, str(count), f"{percent:.2%}")
        console = Console(record=True, file=io.StringIO())
        console.print(table)
        logger.info(console.export_text(styles=True))

        table = Table(title="Expressions")
        table.add_column("Type", justify="left")
        table.add_column("count", justify="right")

        all_exprs = GraphFunctions(*get_graphs(self.exprs)).nodes_of_type(Arithmetic)
        expr_types = groupby(all_exprs, type)
        for expr_type, exprs_for_type in expr_types.items():
            table.add_row(expr_type.__name__, str(len(exprs_for_type)))
        console = Console(record=True, file=io.StringIO())
        console.print(table)
        logger.info(console.export_text(styles=True))

        logger.info(self.times)

    def finish(self):
        self._total.__exit__(None, None, None)
        self.print()
        self.singleton = None

    def event(
        self,
        name: str,
        expr: Arithmetic,
        terminal: bool = True,
        exc: Exception | None = None,
    ):
        if exc:
            name = f"{name} {type(exc).__name__}"
        self.exprs[expr].add(name)
        self.events[name].add(expr)
        self.times.add(name)

        if terminal and len(self.exprs) % 1000 == 0:
            self.print()


@pytest.fixture
def cleanup_stats():
    yield
    Stats.get().finish()


@pytest.mark.slow
@pytest.mark.usefixtures("cleanup_stats")
@given(st_exprs.trees)
@settings(
    deadline=None,
    max_examples=int(NUM_EXAMPLES),
    phases=(
        Phase.generate,
        # Phase.reuse,
        # Phase.explicit,
        Phase.target,
        # Phase.shrink,
        # Phase.explain,
    ),
    suppress_health_check=[
        HealthCheck.data_too_large,
        HealthCheck.too_slow,
        HealthCheck.filter_too_much,
        HealthCheck.large_base_example,
    ],
)
def test_folding_statistics(expr: Arithmetic):
    """
    Run with:
    ```bash
    FBRK_STIMEOUT=5 \
    FBRK_SPARTIAL=n \
    FBRK_SMAX_ITERATIONS=10 \
    ./test/runpytest.sh  \
    -k "test_folding_statistics"
    ```
    """
    stats = Stats.get()
    stats.event("generate", expr, terminal=False)
    solver = DefaultSolver()
    root = Parameter(domain=Numbers(negative=True, zero_allowed=True, integer=False))
    root.alias_is(expr)

    try:
        evaluated_expr = evaluate_expr(expr)
        stats.event("evaluate", expr, terminal=False)
    except NotImplementedError:
        stats.event("not implemented in literals", expr)
        return
    except Exception as e:
        stats.event("eval exc", expr, exc=e)
        return

    assert isinstance(evaluated_expr, Quantity_Interval_Disjoint)

    try:
        solver.update_superset_cache(root)
        solver_result = solver.inspect_get_known_supersets(root)
        assert isinstance(solver_result, Quantity_Interval_Disjoint)
    except Contradiction:
        stats.event("contradiction", expr)
        return
    except TimeoutError:
        stats.event("timeout", expr)
        return
    except NotImplementedError:
        stats.event("not implemented in solver", expr)
        return
    except Exception as e:
        stats.event("exc", expr, exc=e)
        return

    if solver_result != evaluated_expr:
        try:
            deviation = solver_result.op_deviation_to(evaluated_expr, relative=True)
        except Exception:
            deviation = solver_result.op_deviation_to(evaluated_expr)
            if deviation <= 1:
                stats.event("incorrect <= 1 ", expr)
            else:
                stats.event("incorrect > 1 ", expr)
            return
        if deviation < 0.01:
            stats.event("incorrect <1% dev", expr)
        elif deviation < 0.1:
            stats.event("incorrect <10% dev", expr)
        else:
            stats.event("incorrect >10% dev", expr)
        return

    stats.event("correct", expr)


# DEBUG --------------------------------------------------------------------------------


def generate_exprs():
    from rich.console import Console
    from rich.markdown import Markdown

    console = Console()

    console.print(Markdown("# Expressions"))

    for _ in range(50):
        expr = st_exprs.trees.example()
        console.print(Markdown("- " + expr.compact_repr()))


def evaluate_exprs():
    from rich.console import Console
    from rich.table import Table

    table = Table(title="Expression Evaluation Examples")
    table.add_column("Evaluated Result", justify="left")
    table.add_column("Original Expression", justify="left")

    for _ in range(50):
        expr = st_exprs.trees.example()

        try:
            result = evaluate_expr(expr)
        except (NotImplementedError, OverflowError) as e:
            result = repr(e)

        table.add_row(str(result), expr.compact_repr())

    console = Console()
    console.print(table)


def main(target: str):
    warnings.filterwarnings("ignore", category=NonInteractiveExampleWarning)

    match target:
        case "evaluate":
            evaluate_exprs()
        case "generate":
            generate_exprs()
        case _:
            raise ValueError(f"Unknown command: {sys.argv[1]}")


if __name__ == "__main__":
    typer.run(main)
